#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis


"""Loss model objects and associated methods."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from itertools import product
from pathlib import Path
from typing import TYPE_CHECKING, Any, overload

import numpy as np
import pandas as pd
from scipy.interpolate import RegularGridInterpolator

from pelicun import base, file_io, uq
from pelicun.model.demand_model import (
    _assemble_required_demand_data,
    _get_required_demand_type,
    _verify_edps_available,
)
from pelicun.model.pelicun_model import PelicunModel
from pelicun.pelicun_warnings import InconsistentUnitsError

if TYPE_CHECKING:
    from collections.abc import Callable

    from pelicun.assessment import AssessmentBase

idx = base.idx


class LossModel(PelicunModel):
    """
    Manages loss information used in assessments.

    Contains a loss model for components with Damage States (DS) and
    one for components with Loss Functions (LF).

    """

    __slots__ = ['ds_model', 'dv_units', 'lf_model']

    def __init__(
        self,
        assessment: AssessmentBase,
        decision_variables: tuple[str, ...] = ('Cost', 'Time'),
        dv_units: dict[str, str] | None = None,
    ) -> None:
        """
        Initialize LossModel objects.

        Parameters
        ----------
        assessment: pelicun.AssessmentBase
            Parent assessment
        decision_variables: tuple
            Defines the decision variables to be included in the loss
            calculations. Defaults to those supported, but fewer can be
            used if desired. When fewer are used, the loss parameters for
            those not used will not be required.

        """
        super().__init__(assessment)

        self.ds_model: RepairModel_DS = RepairModel_DS(assessment)
        self.lf_model: RepairModel_LF = RepairModel_LF(assessment)
        self._loss_map = None
        self.decision_variables = decision_variables
        self.dv_units = dv_units

    @property
    def sample(self) -> pd.DataFrame | None:
        """
        Combines the samples of the ds_model and lf_model sub-models.

        Returns
        -------
        pd.DataFrame
            The combined loss sample.

        """
        # Handle `None` cases

        if self.ds_model.sample is None and self.lf_model.sample is None:
            return None

        if self.ds_model.sample is None:
            return self.lf_model.sample

        if self.lf_model.sample is None:
            return self.ds_model.sample

        # If both are not None, combine

        ds_model_levels = self.ds_model.sample.columns.names

        # add a `ds` level to the lf_model sample
        new_index = self.lf_model.sample.columns.to_frame(index=False)
        # add
        new_index['ds'] = 'N/A'
        # reorder
        new_index = new_index[ds_model_levels]
        new_multiindex = pd.MultiIndex.from_frame(new_index)
        self.lf_model.sample.columns = new_multiindex

        return pd.concat((self.ds_model.sample, self.lf_model.sample), axis=1)

    @property
    def loss_params(self) -> pd.DataFrame | None:
        """
        Combines the loss params of the ds_model and lf_model sub-models.

        Returns
        -------
        pd.DataFrame
            The combined loss parameter table

        """
        # Handle `None` cases

        if self.ds_model.loss_params is None and self.lf_model.loss_params is None:
            return None

        if self.ds_model.loss_params is None:
            return self.lf_model.loss_params

        if self.lf_model.loss_params is None:
            return self.ds_model.loss_params

        # If both are not None, combine

        return pd.concat((self.ds_model.loss_params, self.lf_model.loss_params))

    @property
    def decision_variables(self) -> tuple[str, ...]:
        """
        Retrieve the decision variables.

        Returns
        -------
        tuple
            Decision variables.

        """
        # pick the object from one of the models
        # it's the same for the other(s).
        return self.ds_model.decision_variables

    @decision_variables.setter
    def decision_variables(self, decision_variables: tuple[str, ...]) -> None:
        """
        Set the decision variables.

        Supported: {`Cost`, `Time`, `Energy`, `Carbon`}.
        Could also be any other string, as long as the provided loss
        parameters contain that decision variable.

        """
        # assign the same DVs to the included loss models.
        for model in self._loss_models:
            model.decision_variables = decision_variables

    def add_loss_map(
        self,
        loss_map_path: str | pd.DataFrame | None = None,
        loss_map_policy: str | None = None,
    ) -> None:
        """
        Add a loss map to the loss model.

        A loss map defines what loss parameter definition should be
        used for each component ID in the asset model.

        Parameters
        ----------
        loss_map_path: str or pd.DataFrame or None
            Path to a csv file or DataFrame object that maps
            components IDs to their loss parameter definitions.
        loss_map_policy: str or None
            If None, does not modify the loss map.
            If set to `fill`, each component ID that is present in
            the asset model but not in the loss map is mapped to
            itself, but `excessiveRID` is excluded.
            If set to `fill_all`, each component ID that is present in
            the asset model but not in the loss map is mapped to
            itself without exceptions.


        Raises
        ------
        ValueError
            If both arguments are None.

        """
        self.log.msg('Loading loss map...')

        # If no loss map is provided and no default is requested,
        # there is no loss map and we can't proceed.
        if loss_map_path is None and loss_map_policy is None:
            msg = 'Please provide a loss map and/or a loss map extension policy.'
            raise ValueError(msg)

        # get a list of unique component IDs
        cmp_set = set(self._asmnt.asset.list_unique_component_ids())

        if loss_map_path is not None:
            self.log.msg('Loss map is provided.', prepend_timestamp=False)
            # Read the loss map into a variable
            loss_map = file_io.load_data(
                loss_map_path,
                None,
                orientation=1,
                reindex=False,
                log=self._asmnt.log,
            )
            assert isinstance(loss_map, pd.DataFrame)
            # <backwards compatibility>
            if np.any(['DMG' in x for x in loss_map.index]):  # type: ignore
                self.log.warning(
                    'The `DMG-` flag in the loss_map index is deprecated '
                    'and no longer necessary. '
                    'Please do not prepend `DMG-` before the component '
                    'names in the loss map.'
                )
                loss_map.index = pd.Index([x[1] for x in loss_map.index])

        else:
            self.log.msg('Using default loss map.', prepend_timestamp=False)
            # Instantiate an empty loss map.
            loss_map = pd.DataFrame({'Repair': pd.Series(dtype='object')})
            loss_map.index = loss_map.index.astype('object')

        if loss_map_policy in {'fill', 'fill_all'}:
            # Populate missing rows with cmp_id -> cmp_id
            for component in cmp_set:
                if component not in loss_map.index:
                    if loss_map_policy == 'fill' and component == 'excessiveRID':
                        continue
                    loss_map.loc[component, :] = component

        elif loss_map_policy is None:
            # Don't do anything.
            pass

        # TODO(AZ): add other loss map policies.
        else:
            msg = f'Unknown loss map policy: `{loss_map_policy}`.'
            raise ValueError(msg)

        # Assign the loss map to the available loss models
        self._loss_map = loss_map

        self.log.msg('Loss map loaded successfully.', prepend_timestamp=True)

    def load_model(
        self,
        data_paths: list[str | pd.DataFrame],
        loss_map: str | pd.DataFrame,
        decision_variables: tuple[str, ...] | None = None,
    ) -> None:
        """<backwards compatibility>."""
        self.log.warning(
            '`load_model` is deprecated and will be dropped in '
            'future versions of pelicun. '
            'Please use `load_model_parameters` instead.'
        )
        self.add_loss_map(loss_map)
        self.load_model_parameters(data_paths, decision_variables)

    def load_model_parameters(
        self,
        data_paths: list[str | pd.DataFrame],
        decision_variables: tuple[str, ...] | None = None,
    ) -> None:
        """
        Load loss model parameters.

        Parameters
        ----------
        data_paths: list of (string | DataFrame)
            List of paths to data or files with loss model
            information. Default XY datasets can be accessed as
            PelicunDefault/XY. Order matters. Parameters defined in
            prior elements in the list take precedence over the same
            parameters in subsequent data paths. I.e., place the
            Default datasets in the back.
        decision_variables: tuple
            Defines the decision variables to be included in the loss
            calculations. Defaults to those supported, but fewer can be
            used if desired. When fewer are used, the loss parameters for
            those not used will not be required.

        """
        if decision_variables is not None:
            # <backwards compatibility>
            self.decision_variables = decision_variables
            self.log.warning(
                'The `decision_variables` argument has been removed. '
                'Please set your desired decision variables like so: '
                '{assessment object}.loss.decision_variables '
                "= ('dv1', 'dv2', ...) before calling "
                '{assessment object}.add_loss_map().'
            )

        self.log.div()
        self.log.msg('Loading loss parameters...')

        # replace `PelicunDefault/` flag with default data path
        data_paths = file_io.substitute_default_path(data_paths, log=self.log)

        #
        # load loss parameter data into the models
        #

        for data_path in data_paths:
            self._load_from_data_path(data_path)

        self.log.msg(
            'Loss model parameters loaded successfully.', prepend_timestamp=False
        )

        #
        # remove items
        #

        self.log.msg(
            'Removing unused loss model parameters.', prepend_timestamp=False
        )

        assert self._loss_map is not None
        for loss_model in self._loss_models:
            # drop unused loss parameter definitions
            loss_model.drop_unused_loss_parameters(self._loss_map)
            # remove components with incomplete loss parameters
            loss_model.remove_incomplete_components()

        # drop unused damage state columns
        self.ds_model.drop_unused_damage_states()

        #
        # obtain DV units
        #
        dv_units: dict = {}
        if self.ds_model.loss_params is not None:
            dv_units.update(
                self.ds_model.loss_params['DV', 'Unit']
                .groupby(level=[1])
                .first()
                .to_dict()
            )
        if self.lf_model.loss_params is not None:
            dv_units.update(
                self.lf_model.loss_params['DV', 'Unit']
                .groupby(level=[1])
                .first()
                .to_dict()
            )
        self.dv_units = dv_units

        #
        # convert units
        #

        self.log.msg(
            'Converting loss model parameter units.', prepend_timestamp=False
        )
        for loss_model in self._loss_models:
            loss_model.convert_loss_parameter_units()

        #
        # verify loss parameter availability
        #

        self.log.msg(
            'Checking loss model parameter '
            'availability for all components in the asset model.',
            prepend_timestamp=False,
        )
        self._ensure_loss_parameter_availability()

    def _load_from_data_path(self, data_path: str | pd.DataFrame) -> None:
        if 'bldg_repair_DB' in data_path:
            data_path = data_path.replace('bldg_repair_DB', 'loss_repair_DB')
            self.log.warning(
                '`bldg_repair_DB` is deprecated and will '
                'be dropped in future versions of pelicun. '
                'Please use `loss_repair_DB` instead.'
            )
        data = file_io.load_data(
            data_path, None, orientation=1, reindex=False, log=self._asmnt.log
        )
        assert isinstance(data, pd.DataFrame)

        # Check for unit consistency
        data.index.names = ['cmp', 'dv']
        units_isolated = data.reset_index()[[('dv', ''), ('DV', 'Unit')]]
        units_isolated.columns = pd.Index(['dv', 'Units'])
        units_isolated_grp = units_isolated.groupby('dv')['Units']
        unit_counts = units_isolated_grp.nunique()
        more_than_one = unit_counts[unit_counts > 1]
        if not more_than_one.empty:
            raise InconsistentUnitsError

        # determine if the loss model parameters are for damage
        # states or loss functions
        if _is_for_ds_model(data):
            self.ds_model.load_model_parameters(data)
        elif _is_for_lf_model(data):
            self.lf_model.load_model_parameters(data)
        else:
            msg = f'Invalid loss model parameters: {data_path}'
            raise ValueError(msg)

    def calculate(self) -> None:
        """
        Calculate the loss of each component block.

        Note: This method simply calculates the loss of each component
        block without any special treatment to `replacement`
        consequences. This can be done at a later step with the
        `aggregate_losses` method.

        Raises
        ------
        ValueError
            If the size of the demand sample and the damage sample
            don't match.

        """
        self.log.div()
        self.log.msg('Calculating losses...')

        # Get the damaged quantities in each damage state for each
        # component of interest.
        demand = self._asmnt.demand.sample
        assert demand is not None
        demand_offset = self._asmnt.options.demand_offset
        assert demand_offset is not None
        nondirectional_multipliers = self._asmnt.options.nondir_multi_dict
        assert nondirectional_multipliers is not None
        assert self._asmnt.asset.cmp_sample is not None
        cmp_sample = self._asmnt.asset.cmp_sample.to_dict('series')
        cmp_marginal_params = self._asmnt.asset.cmp_marginal_params
        assert cmp_marginal_params is not None
        if self._asmnt.damage.ds_model.sample is not None:
            # TODO(JVM): FIND A WAY to avoid making a copy of this.
            dmg_quantities = self._asmnt.damage.ds_model.sample.copy()
            if len(demand) != len(dmg_quantities):
                msg = (
                    f'The demand sample contains {len(demand)} realizations, '
                    f'but the damage sample contains {len(dmg_quantities)}. '
                    f'Loss calculation cannot proceed when '
                    f'these numbers are different. '
                )
                raise ValueError(msg)
            self.ds_model.calculate(dmg_quantities)

        self.lf_model.calculate(
            demand,
            cmp_sample,
            cmp_marginal_params,
            demand_offset,
            nondirectional_multipliers,
        )

        self.log.msg('Loss calculation successful.')

    def consequence_scaling(self, scaling_specification: str) -> None:
        """
        Apply scale factors to losses.

        Applies scale factors to the loss sample according to the
        given scaling specification. The scaling specification should
        be a path to a CSV file. It should contain a `Decision
        Variable` column with a specified decision variable in each
        row. Other optional columns are `Component`, `Location`,
        `Direction`. Each row acts as an independent scaling
        operation, with the scale factor defined in the `Scale Factor`
        column, which is required. If any value is missing in the
        optional columns, it is assumed that the scale factor should
        be applied to all entries of the loss sample where the other
        column values match. For example, if the specification has a
        single row with `Decision Variable` set to 'Cost', `Scale
        Factor` set to 2.0, and no other columns, this will double the
        'Cost' DV. If instead `Location` was also set to `1`, it would
        double the Cost of all components that have that location. The
        columns `Location` and `Direction` can contain ranges, like
        this: `1--3` means `1`, `2`, and `3`. If a range is used in
        both `Location` and `Direction`, the factor of that row will
        be applied once to all combinations.

        Parameters
        ----------
        scaling_specification: str
            Path to a CSV file containing the scaling specification.

        Raises
        ------
        ValueError
            If required columns are missing or contain NaNs.

        """
        # Specify expected dtypes from the start.
        dtypes = {
            'Decision Variable': 'str',
            'Component': 'str',
            'Location': 'str',
            'Direction': 'str',
            'Scale Factor': 'float64',
        }

        scaling_specification_df = pd.read_csv(scaling_specification, dtype=dtypes)

        if (
            'Decision Variable' not in scaling_specification_df.columns
            or scaling_specification_df['Decision Variable'].isna().any()
        ):
            msg = (
                'The `Decision Variable` column is missing '
                'from the scaling specification or contains NaN values.'
            )
            raise ValueError(msg)
        if (
            'Scale Factor' not in scaling_specification_df.columns
            or scaling_specification_df['Scale Factor'].isna().any()
        ):
            msg = (
                'The `Scale Factor` column is missing '
                'from the scaling specification or contains NaN values.'
            )
            raise ValueError(msg)

        # Add missing optional columns with NaN values
        optional_cols = ['Component', 'Location', 'Direction']
        for col in optional_cols:
            if col not in scaling_specification_df.columns:
                scaling_specification_df[col] = np.nan

        # Rename the columns to the internally used values
        name_map = {
            'Decision Variable': 'dv',
            'Component': 'dmg',
            'Location': 'loc',
            'Direction': 'dir',
            'Scale Factor': 'scaling',
        }
        scaling_specification_df = scaling_specification_df.rename(columns=name_map)

        # Expand ranges in 'loc' and 'dir'
        def _expand_range(col):  # noqa: ANN001, ANN202
            if pd.isna(col):
                return [col]
            if '--' in col:
                start, end = (int(x) for x in col.split('--'))
                return [str(x) for x in range(start, end + 1)]
            return [col]

        # Generate all combinations of loc and dir ranges
        expanded_df = scaling_specification_df.apply(  # type: ignore
            lambda row: pd.DataFrame(
                list(product(_expand_range(row['loc']), _expand_range(row['dir']))),
                columns=['loc', 'dir'],
            ).assign(dv=row['dv'], dmg=row['dmg'], scaling=row['scaling']),
            axis=1,
        )

        expanded_df = pd.concat(expanded_df.values)

        # Now, for each unique combination in expanded_df, apply
        # consequence scaling
        for _, row in expanded_df.iterrows():
            scaling_conditions = {
                k: row[k] for k in ('dv', 'dmg', 'loc', 'dir') if not pd.isna(row[k])
            }
            self._apply_consequence_scaling(
                scaling_conditions, row['scaling'], raise_missing=False
            )

    def _apply_consequence_scaling(
        self,
        scaling_conditions: dict,
        scale_factor: float,
        *,
        raise_missing: bool = True,
    ) -> None:
        """
        Apply a scale factor to selected loss sample columns.

        The scaling conditions are passed as a dictionary mapping
        level names with their required value for the condition to be
        met. It has to contain `dv` as one of its keys, defining the
        decision variable where the factors should be applied. Other
        valid levels include:
        - `dmg`: containing a source component name,
        - `loc`: containing a location,
        - `dir`: containing a direction,
        - `uid`: containing a Unique Component ID (UID).

        If any of the keys is missing, it is assumed that the scaling
        factor should be applied to all relevant consequences (those
        matching the remaining values of the hierarchical index).

        Parameters
        ----------
        scaling_conditions: dict
            A dictionary mapping level names with a single value. Only the
            rows where the index levels have the provided values will be
            affected. The dictionary can be empty, in which case all rows
            will be affected, or contain only some levels and values, in
            which case only the matching rows will be affected.
        scale_factor: float
            Scale factor to use.
        raise_missing: bool
            Raise an error if no rows are matching the given conditions.

        Raises
        ------
        ValueError
            If the scaling_conditions dictionary does not contain a
            `dv` key.

        """
        # make sure we won't apply the same factor to all DVs at once,
        # highly unlikely anyone would actually want to do this.
        if 'dv' not in scaling_conditions:
            msg = (
                'The index of the `scaling_conditions` dictionary '
                'should contain a level named `dv` listing the '
                'relevant decision variable.'
            )
            raise ValueError(msg)

        for model in self._loss_models:
            # check if it's empty
            if model.sample is None:
                continue

            # ensure the levels exist (but don't check if specified
            # values exist yet)
            for name in scaling_conditions:
                if name not in model.sample.columns.names:
                    msg = (
                        f'`scaling_conditions` contains an unknown level: `{name}`.'
                    )
                    raise ValueError(msg)

            # apply scale factors
            base.multiply_factor_multiple_levels(
                model.sample,
                scaling_conditions,
                scale_factor,
                axis=1,
                raise_missing=raise_missing,
            )

    def save_sample(
        self, filepath: str | None = None, *, save_units: bool = False
    ) -> None | pd.DataFrame | tuple[pd.DataFrame, pd.Series]:
        """
        Save a combined sample of the `ds_model` and `lf_model`.

        This method handles the storage of a sample of loss estimates,
        which can either be saved directly to a file or returned as a
        DataFrame for further manipulation. When saving to a file,
        additional information such as unit conversion factors and
        column units can be included. If the data is not being saved
        to a file, the method can return the DataFrame with or without
        units as specified.

        Parameters
        ----------
        filepath: str, optional
            The path to the file where the loss sample should be
            saved. If not provided, the sample is not saved to disk
            but returned.
        save_units: bool, default: False
            Indicates whether to include a row with unit information
            in the returned DataFrame. This parameter is ignored if a
            file path is provided.

        Returns
        -------
        None or tuple
            If `filepath` is provided, the function returns None after
            saving the data.
            If no `filepath` is specified, returns:
            * DataFrame containing the loss sample.
            * Optionally, a Series containing the units for each
            column if `save_units` is True.

        """
        self.log.div()
        if filepath is not None:
            self.log.msg('Saving loss sample...')

        assert self.sample is not None
        assert self.loss_params is not None

        cmp_units = self.loss_params['DV', 'Unit'].sort_index()
        dv_units = pd.Series(
            index=self.sample.columns, name='Units', dtype='object'
        ).sort_index()

        valid_dv_types = dv_units.index.unique(level=0)
        valid_cmp_ids = dv_units.index.unique(level=1)

        for cmp_id, dv_type in cmp_units.index:
            if (dv_type in valid_dv_types) and (cmp_id in valid_cmp_ids):
                dv_units.loc[dv_type, cmp_id] = cmp_units.loc[cmp_id, dv_type]

        res = file_io.save_to_csv(
            self.sample,
            Path(filepath) if filepath is not None else None,
            units=dv_units,
            unit_conversion_factors=self._asmnt.unit_conversion_factors,
            use_simpleindex=(filepath is not None),
            log=self._asmnt.log,
        )
        if filepath is not None:
            self.log.msg('Loss sample successfully saved.', prepend_timestamp=False)
            return None

        assert isinstance(res, pd.DataFrame)

        units = res.loc['Units']
        res = res.drop('Units')
        res = res.astype(float)
        assert isinstance(res, pd.DataFrame)
        assert isinstance(units, pd.Series)

        if save_units:
            return res, units

        return res

    def load_sample(self, filepath: str | pd.DataFrame) -> None:
        """
        <backwards compatibility>.

        Saves the sample of the `ds_model`.

        """
        self.log.warning(
            '`{loss model}.load_sample` is deprecated and will raise '
            'in future versions of pelicun. Please use '
            '{loss model}.ds_model.load_sample instead.'
        )
        dv_units = self.ds_model.load_sample(filepath=filepath)
        self.dv_units = dv_units

    def aggregate_losses(  # noqa: C901
        self,
        replacement_configuration: (
            tuple[uq.RandomVariableRegistry, dict[str, float]] | None
        ) = None,
        loss_combination: dict | None = None,
        *,
        future: bool = False,
    ) -> pd.DataFrame | tuple[pd.DataFrame, pd.DataFrame]:
        """
        Aggregate the losses produced by each component.

        Parameters
        ----------
        replacement_configuration: Tuple, optional
            Tuple containing a RandomVariableRegistry and a
            dictionary. The RandomVariableRegistry is defining
            building replacement consequence RVs for the active
            decision variables. The dictionary defines exceedance
            thresholds. If the aggregated value for a decision
            variable (conditioned on no replacement) exceeds the
            threshold, then replacement is triggered. This can happen
            for multiple decision variables at the same
            realization. The consequence keyword `replacement` is
            reserved to represent exclusive triggering of the
            replacement consequences, and other consequences are
            ignored for those realizations where replacement is
            triggered. When assigned to None, then `replacement` is
            still treated as an exclusive consequence (other
            consequences are set to zero when replacement is nonzero)
            but it is not being additionally triggered by the
            exceedance of any thresholds. The aggregated loss sample
            contains an additional column with information on whether
            replacement was already present or triggered by a
            threshold exceedance for each realization.
        loss_combination: dict, optional
            Dictionary defining how losses for specific components
            should be aggregated for a given decision variable. It has
            the following structure: {`dv`: {(`c1`, `c2`): `arr`,
            ...}, ...}, where `dv` is some decision variable, (`c1`,
            `c2`) is a tuple defining a component pair, `arr` is a NxN
            numpy array defining a combination table, and `...` means
            that more key-value pairs with the same schema can exist
            in the dictionaries.  The loss sample is expected to
            contain columns that include both `c1` and `c2` listed as
            the component. The combination is applied to all pairs of
            columns where the components are `c1` and `c2`, and all of
            the rest of the multiindex levels match (`loc`, `dir`,
            `uid`). This means, for example, that when combining wind
            and flood losses, the asset model should contain both a
            wind and a flood component defined at the same
            location-direction.  `arr` can also be an M-dimensional
            numpy array where each dimension has length N (NxNx...xN).
            This structure allows for the loss combination of M
            components.  In this case the (`c1`, `c2`) tuple should
            contain M elements instead of two.
        future: bool, optional
            Defaults to False. When set to True, it enables the
            updated return type.

        Notes
        -----
        Regardless of the value of the arguments, this method does not
        alter the state of the loss model, i.e., it does not modify
        the values of the `.sample` attributes.

        Returns
        -------
        dataframe or tuple
            Dataframe with the aggregated loss of each realization,
            and another boolean dataframe with information on which DV
            thresholds were exceeded in each realization, triggering
            replacement. If no thresholds are specified it only
            contains False values. The second dataframe is only
            returned with `future` set to True.

        """
        # TODO(JVM): When we start working on the documentation,
        # simplify the docstring above and point the relevant detailed
        # section in the documentation.

        # validate input
        if replacement_configuration is not None:
            self._validate_input_replacement_thresholds(replacement_configuration)
        # validate loss_combination input
        if loss_combination is not None:
            self._validate_input_loss_combination(loss_combination)

        #
        # operate on copes of the loss samples to avoid altering them.
        #

        if self.ds_model.sample is not None:
            ds_sample = self.ds_model.sample.copy()
        else:
            ds_sample = None
        if self.lf_model.sample is not None:
            lf_sample = self.lf_model.sample.copy()
        else:
            lf_sample = None

        def _construct_columns() -> list[str]:
            columns = [
                f'repair_{x.lower()}' for x in self.decision_variables if x != 'Time'
            ]
            # Note: The `Time` DV gets special treatment.
            # create the summary DF
            if 'Time' in self.decision_variables:
                columns.extend(('repair_time-sequential', 'repair_time-parallel'))
            return columns

        if ds_sample is None and lf_sample is None:
            self.log.msg('There are no losses.')
            return pd.DataFrame(0.00, index=[0], columns=_construct_columns())

        #
        #  handle `replacement`, regardless of whether
        # `replacement_thresholds` is empty. (if `replacement`
        # occurs, we ignore the losses from other componnets)
        #

        if ds_sample is not None:
            self._make_replacement_exclusive(ds_sample, lf_sample)

        #
        # combine samples
        #

        # levels to preserve (this aggregates `ds` for the ds_model)
        column_levels = ['dv', 'loss', 'dmg', 'loc', 'dir', 'uid']
        combined_sample = self.sample
        sample = (
            combined_sample.groupby(level=column_levels, axis=1)  # type: ignore
            .sum()
            .sort_index(axis=1)
        )

        #
        # perform loss combinations (special non-additive
        # aggregations, e.g., Wind + Flood)
        #

        if loss_combination is not None:
            sample = self._apply_loss_combinations(loss_combination, sample)

        #
        # consider replacement threshold values
        #

        sample, exceedance_bool_df = self._apply_replacement_thresholds(
            sample, replacement_configuration
        )

        # Sum-up component losses
        df_agg = self._aggregate_sample(sample, _construct_columns())

        if not future:
            self.log.warning(
                '`aggregate_losses` has been expanded to support the '
                'consideration of the exceedance of loss threshold '
                'values leading to asset replacement '
                '(like excessive repair costs). The new implementation '
                'returns a tuple where the first element is the '
                'aggregated losses and the second contains information '
                'on which decision variables triggered replacement '
                'considering the specified replacement trhesholds. '
                'To obtain the new output and silence this warning, '
                'please specify `future=True` as an argument to this method.'
            )
            return df_agg

        return df_agg, exceedance_bool_df

    def _validate_input_loss_combination(self, loss_combination: dict) -> None:
        for dv, combinations in loss_combination.items():
            if dv not in self.decision_variables:
                msg = (
                    f'`loss_combination` contains the key '
                    f'`{dv}` which is not found in the active '
                    f'decision variables. These are: '
                    f'{self.decision_variables}.'
                )
                raise ValueError(msg)
            for components, array in combinations.items():
                if not isinstance(components, tuple):
                    msg = (
                        f'Invalid type for components in loss combination '
                        f'for `{dv}`: {type(components)}. It should be a tuple.'
                    )
                    raise TypeError(msg)
                if not all(isinstance(c, str) for c in components):
                    msg = (
                        f'All elements of the components tuple in loss '
                        f'combination for `{dv}` should be strings.'
                    )
                    raise ValueError(msg)
                if not isinstance(array, np.ndarray):
                    msg = (
                        f'Invalid type for array in loss combination '
                        f'for `{dv}`: {type(array)}. It should be a numpy array.'
                    )
                    raise TypeError(msg)

    def _validate_input_replacement_thresholds(
        self,
        replacement_configuration: tuple[
            uq.RandomVariableRegistry, dict[str, float]
        ],
    ) -> None:
        replacement_consequence_rv_reg, replacement_ratios = (
            replacement_configuration
        )
        if not isinstance(replacement_consequence_rv_reg, uq.RandomVariableRegistry):
            msg = (
                f'Invalid type for replacement consequence RV registry: '
                f'{type(replacement_consequence_rv_reg)}. It should be '
                f'uq.RandomVariableRegistry.'
            )
            raise TypeError(msg)
        for key in replacement_consequence_rv_reg.RV:
            if key not in self.decision_variables:
                msg = (
                    f'`replacement_consequence_RV_reg` contains the key '
                    f'`{key}` which is not found in the active '
                    f'decision variables. These are: '
                    f'{self.decision_variables}.'
                )
                if self.query_error_setup(
                    'Loss/ReplacementThreshold/RaiseOnUnknownKeys'
                ):
                    raise ValueError(msg)
                self.log.warning(msg)

        for key in replacement_ratios:
            if key not in self.decision_variables:
                msg = (
                    f'`replacement_ratios` contains the key '
                    f'`{key}` which is not found in the active '
                    f'decision variables. These are: '
                    f'{self.decision_variables}.'
                )
                if self.query_error_setup(
                    'Loss/ReplacementThreshold/RaiseOnUnknownKeys'
                ):
                    raise ValueError(msg)
                self.log.warning(msg)
        # The replacement_consequence_RV_reg should contain an RV for
        # all active DVs, regardless of whether there is a replacement
        # threshold for that DV, because when replacememnt is
        # triggered, we need to assign a consequence for all DVs.
        for key in self.decision_variables:
            if key not in replacement_consequence_rv_reg.RV:
                msg = f'Missing replacement consequence RV ' f'for `{key}`.'
                raise ValueError(msg)

    def _apply_loss_combinations(
        self, loss_combination: dict, sample: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Perform loss combinations of specified components.

        This function deconstructs the loss combination arrays,
        identifies the combinable components, and applies the
        specified loss combinations to the sample data. The
        transformed sample, including the combined columns, is
        returned as a new DataFrame.

        Parameters
        ----------
        loss_combination: dict
            A dictionary containing the loss combination
            information. The structure is nested dictionaries where
            the outer keys are decision variables, inner keys are
            components to combine, and the values are array objects
            representing the combination data.

        sample: pandas.DataFrame
            The input DataFrame containing the sample data. The
            columns are assumed to be a MultiIndex with at least the
            levels (decision_variable, loss_id, component_id,
            location, direction, uid).

        Returns
        -------
        pandas.DataFrame
            A new DataFrame with the combined loss data.

        """
        # deconstruct combination arrays to extract the input domains
        loss_combination_converted = self._deconstruct_loss_combination_arrays(
            loss_combination
        )

        # initialize variables

        # sample as dictionary for fast column retrieval
        dsample = {col: sample[col] for col in sample.columns}

        # Transformed sample (includes the combined columns), as
        # dictionary. Will be turned into a dataframe in the end.
        # This avoids manipulating the original sample dataframe which
        # would be slow.
        dcsample: dict = {}

        # add columns to the new sample dictionary.
        # those that should be combined
        self._loss_combination_add_combinable(
            dsample, loss_combination_converted, dcsample
        )
        # and the remaining
        for col, val in dsample.items():
            dcsample[col] = val  # noqa: PERF403

        # turn into a dataframe
        return pd.DataFrame(dcsample).rename_axis(columns=sample.columns.names)

    def _loss_combination_add_combinable(
        self, dsample: dict, loss_combination_converted: dict, dcsample: dict
    ) -> None:
        """
        Add combinable loss data.

        This function identifies groups of `loc`-`dir`-`uid` that can
        be combined for each decision variable and computes the
        combined loss using interpolation functions. It modifies the
        given datasets `dsample` and `dcsample` in-place, removing
        combinable columns from dsample and adding the combined losses
        to dcsample.

        Parameters
        ----------
        dsample: dict
            A dictionary representing the loss sample data, where keys
            are tuples of the form (decision_variable, loss_id,
            component_id, location, direction, uid) and values are the
            corresponding data arrays.

        loss_combination_converted: dict
            A dictionary containing loss combination data. The
            structure is nested dictionaries where the outer keys are
            decision variables, inner keys are components to combine,
            and the values are tuples of combination parameters
            (domains and reference values).

        dcsample: dict
            A dictionary to store the combined loss data, where keys
            are tuples of the form (decision_variable, 'combination',
            combined_component_string, location, direction, uid) and
            values are the combined loss data arrays.

        """
        dmg_to_loss = self._map_component_ids_to_loss_ids(dsample)

        # identify all `loc`-`dir`-`uid`s that can be grouped for each
        # decision variable.
        potential_groups = self._identify_potential_groups(dsample)

        # cache already defined interpolation functions. This obviates
        # the need to define all of them and we can just define them
        # on the spot when needed, and reuse them if available.
        interpolation_function_cache: dict = {}

        for (
            decision_variable,
            combination_data,
        ) in loss_combination_converted.items():
            for (
                components_to_combine,
                combination_parameters,
            ) in combination_data.items():
                # determine if the components to combine are part of
                # an available group
                target_group: dict | None = None
                for available_group in potential_groups[decision_variable]:
                    # check if `components_to_combine` is a subset of
                    # that available group
                    if frozenset(components_to_combine) <= available_group:
                        target_group = available_group
                        break
                assert target_group is not None
                # construct relevant loss sample columns
                for loc_dir_uid in potential_groups[decision_variable][target_group]:
                    cols = [
                        (decision_variable, dmg_to_loss[x], x, *loc_dir_uid)
                        for x in target_group
                    ]
                    values = np.column_stack([dsample[col] for col in cols])
                    # define/get interpolation function
                    if (
                        interpolation_function_cache.get(components_to_combine)
                        is not None
                    ):
                        interp_func = interpolation_function_cache.get(
                            components_to_combine
                        )
                    else:
                        domains, reference_values = combination_parameters
                        interp_func = RegularGridInterpolator(
                            domains, reference_values
                        )
                    assert interp_func is not None
                    combined_loss = interp_func(values)
                    combined_loss_col = (
                        decision_variable,
                        'combination',
                        '(' + ', '.join(components_to_combine) + ')',
                        *loc_dir_uid,
                    )
                    dcsample[combined_loss_col] = combined_loss
                    for col in cols:
                        dsample.pop(col)

    def _identify_potential_groups(self, dsample: dict) -> dict:  # noqa: PLR6301
        """
        Identify potential groups of `loc`-`dir`-`uid` for each DV.

        This function identifies all combinations of `loc`-`dir`-`uid`
        that can be grouped for each decision variable based on the
        provided data sample.

        Parameters
        ----------
        dsample: iterable
            An iterable where each containing tuple contains
            information about the components and their attributes. The
            expected format of each tuple is (decision_variable,
            loss_id, component_id, location, direction, uid).

        Returns
        -------
        dict
            A dictionary where keys are decision variables and values
            are nested dictionaries. The nested dictionaries map
            frozensets of component IDs to lists of (location,
            direction, uid) tuples.

        """
        grouped: defaultdict = defaultdict(defaultdict(list).copy)
        for col in dsample:
            c_dv, _, c_dmg, c_loc, c_dir, c_uid = col
            grouped[c_dv][c_loc, c_dir, c_uid].append(c_dmg)
        # invert so that we have component sets mapped to
        # `loc`-`dir`-`uid`s.
        inverted: defaultdict = defaultdict(defaultdict(list).copy)
        for c_dv in grouped:
            for loc_dir_uid, component_set in grouped[c_dv].items():
                inverted[c_dv][frozenset(component_set)].append(loc_dir_uid)
        return inverted

    def _map_component_ids_to_loss_ids(self, dsample: dict) -> dict:  # noqa: PLR6301
        """
        Map component IDs to loss IDs.

        This function maps components to losses based on the loss
        sample's columns. It assumes that multiple component IDs can
        have the same loss ID, but the same component ID cannot have
        multiple loss IDs.

        Parameters
        ----------
        dsample: tuple dictionary keys
            Each tuple contains information about the components and
            corresponding losses.

        Returns
        -------
        dict
            A dictionary where keys are component IDs and values are
            loss IDs.

        """
        dmg_to_loss = {}
        for col in dsample:
            c_loss = col[1]
            c_dmg = col[2]
            dmg_to_loss[c_dmg] = c_loss
        return dmg_to_loss

    def _deconstruct_loss_combination_arrays(self, loss_combination: dict) -> dict:  # noqa: PLR6301
        """
        Deconstruct loss combination arrays.

        This function converts a nested dictionary of loss combination
        arrays into a format suitable for further processing. It
        extracts the array values and the index information from each
        array.

        Parameters
        ----------
        loss_combination: dict
           A dictionary where keys are decision variables and values
           are another dictionary. The inner dictionary has keys as
           components to combine and values as numpy array
           objects representing the combination data.

        Returns
        -------
        dict
            A dictionary with the same structure as the input
            `loss_combination`.  For each decision variable and
            component combination, the array is replaced with a
            tuple containing the combination domain and the combination
            array itself.

        """
        loss_combination_converted: dict = {}
        for decision_variable, combination_data in loss_combination.items():
            loss_combination_converted[decision_variable] = {}
            for (
                components_to_combine,
                combination_array,
            ) in combination_data.items():
                combination_index = (
                    combination_array[:, 0],
                    combination_array[0, :],
                )
                loss_combination_converted[decision_variable][
                    components_to_combine
                ] = (
                    combination_index,
                    combination_array,
                )
        return loss_combination_converted

    def _aggregate_sample(self, sample: pd.DataFrame, columns: list) -> pd.DataFrame:
        """
        Sum up component losses.

        Returns
        -------
        pd.DataFrame
            Dataframe with the aggregated losses.

        """
        df_agg = pd.DataFrame(index=sample.index, columns=columns)
        # group results by DV type and location
        aggregated = sample.groupby(
            level=['dv', 'loc'],
            axis=1,  # type: ignore
        ).sum()

        for decision_variable in self.decision_variables:
            # Time
            if (
                decision_variable == 'Time'
                and 'Time' in aggregated.columns.get_level_values('dv')
            ):
                df_agg['repair_time-sequential'] = aggregated['Time'].sum(axis=1)

                df_agg['repair_time-parallel'] = aggregated['Time'].max(axis=1)
            elif (
                decision_variable == 'Time'
                and 'Time' not in aggregated.columns.get_level_values('dv')
            ):
                df_agg = df_agg.drop(
                    ['repair_time-parallel', 'repair_time-sequential'], axis=1
                )
            # All other
            elif decision_variable in aggregated.columns.get_level_values('dv'):
                df_agg[f'repair_{decision_variable.lower()}'] = aggregated[
                    decision_variable
                ].sum(axis=1)
            else:
                df_agg = df_agg.drop(f'repair_{decision_variable.lower()}', axis=1)

        # Convert units ..
        column_measures = [
            x.replace('repair_', '')
            .replace('-sequential', '')
            .replace('-parallel', '')
            for x in df_agg.columns.get_level_values(0)
        ]
        assert self.dv_units is not None
        column_units = [self.dv_units[x.title()] for x in column_measures]
        dv_units = pd.Series(column_units, index=df_agg.columns, name='Units')
        res = file_io.save_to_csv(
            df_agg,
            None,
            units=dv_units,
            unit_conversion_factors=self._asmnt.unit_conversion_factors,
            use_simpleindex=False,
            log=self._asmnt.log,
        )
        assert isinstance(res, pd.DataFrame)
        df_agg = res
        df_agg = df_agg.drop('Units')
        df_agg = df_agg.astype(float)

        df_agg_mi = base.convert_to_MultiIndex(df_agg, axis=1)
        assert isinstance(df_agg_mi, pd.DataFrame)
        df_agg = df_agg_mi
        df_agg = df_agg.sort_index(axis=1)
        df_agg = df_agg.reset_index(drop=True)
        assert isinstance(df_agg, pd.DataFrame)
        return df_agg

    def _apply_replacement_thresholds(  # noqa: PLR6301
        self,
        sample: pd.DataFrame,
        replacement_configuration: (
            tuple[uq.RandomVariableRegistry, dict[str, float]] | None
        ),
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        # If there is no `replacement_configuration`, simply return.
        if replacement_configuration is None:
            # `exceedance_bool_df` is empty in this case.
            exceedance_bool_df = pd.DataFrame(index=sample.index, dtype=bool)
            return sample, exceedance_bool_df

        replacement_consequence_rv_reg, replacement_ratios = (
            replacement_configuration
        )
        exceedance_bool_df = pd.DataFrame(  # type: ignore
            data=False,
            index=sample.index,
            columns=replacement_consequence_rv_reg.RV.keys(),
            dtype=bool,
        )

        # Sample replacement consequences from the registry
        replacement_consequence_rv_reg.generate_sample(len(sample), 'MonteCarlo')

        sample_dvs = replacement_consequence_rv_reg.RV.keys()
        for sample_dv in sample_dvs:
            sub_sample = sample.loc[:, sample_dv]
            if 'replacement' in sub_sample.columns.get_level_values('loss'):
                # If `replacement` already exists as a consequence,
                # determine the realizations where it is non-zero.
                no_replacement_mask = (
                    ~(sub_sample['replacement'] > 0.00).any(axis=1).to_numpy()
                )
                no_replacement_columns = (
                    sub_sample.columns.get_level_values('loss') != 'replacement'
                )
            else:
                # Otherwise there is no row where we already have replacement
                no_replacement_mask = np.full(len(sub_sample), fill_value=True)
                no_replacement_columns = np.full(
                    len(sub_sample.columns), fill_value=True
                )
            # Get the sum to compare with the thresholds
            consequence_sum_given_no_replacement = sub_sample.iloc[  # type: ignore
                no_replacement_mask, no_replacement_columns
            ].sum(axis=1)
            if not consequence_sum_given_no_replacement.empty:
                consequence_values = replacement_consequence_rv_reg.RV[
                    sample_dv
                ].sample
                assert consequence_values is not None
                exceedance_mask = (
                    consequence_sum_given_no_replacement
                    > consequence_values[no_replacement_mask]
                    * replacement_ratios[sample_dv]
                )
                # Fill the remaining rows with False
                if len(exceedance_mask) < len(sub_sample):
                    exceedance_mask = exceedance_mask.reindex(
                        sub_sample.index, fill_value=False
                    )
            else:
                exceedance_mask = pd.Series(
                    np.full(len(sub_sample), fill_value=False),
                    index=sub_sample.index,
                )

            # Monitor triggering of replacement
            exceedance_bool_df[sample_dv] = exceedance_mask

        # Assign replacement consequences where the threshold has been
        # exceeded.
        exceedance_realizations = exceedance_bool_df.any(axis=1)
        # Assign replacement consequences: needs to include all DVs
        for other_dv in replacement_consequence_rv_reg.RV:
            col = (
                other_dv,
                'replacement',
                'threshold_exceedance',
                '0',
                '1',
                '0',
            )
            # If it doesn't exist, initialize and set to 0.00
            if col not in sample:
                sample[col] = 0.00
                sample = sample.sort_index(axis=1)
            # Assign replacement consequences
            other_sample = replacement_consequence_rv_reg.RV[other_dv].sample
            assert other_sample is not None
            sample.loc[exceedance_realizations, col] = other_sample[
                exceedance_realizations
            ]
        # Remove all other realized consequences from the realizations
        # where the threshold was exceeded.
        sample.loc[
            exceedance_realizations,
            sample.columns.get_level_values('dmg') != 'threshold_exceedance',
        ] = 0.00

        return sample, exceedance_bool_df

    def _make_replacement_exclusive(  # noqa: PLR6301
        self, ds_sample: pd.DataFrame, lf_sample: pd.DataFrame | None
    ) -> None:
        """
        Make the replacement consequence exclusive.

        If `replacement` columns exist in `ds_sample`, this method
        treats all nonzero loss values driven by `replacement` as
        exclusive and zeroes-out the loss values of all other columns
        for the applicable rows.
        """
        # rows where replacement is non-zero
        replacement_rows: list = []

        # columns that correspond to the replacement consequence
        replacement_columns = (
            ds_sample.columns.get_level_values('loss') == 'replacement'
        )
        rows_df = ds_sample.iloc[:, replacement_columns]

        if not rows_df.empty:
            replacement_rows = (
                np.argwhere(np.any(rows_df.to_numpy() > 0.0, axis=1))
                .reshape(-1)
                .tolist()
            )
        ds_sample.iloc[replacement_rows, ~replacement_columns] = 0.00  # type: ignore
        if lf_sample is not None:
            lf_sample.iloc[replacement_rows, :] = 0.00

    @property
    def _loss_models(self) -> tuple[RepairModel_DS, RepairModel_LF]:
        return (self.ds_model, self.lf_model)

    @property
    def _loss_map(self) -> pd.DataFrame | None:
        """
        Returns the loss map.

        Returns
        -------
        pd.DataFrame
            The loss map.

        """
        # Retrieve the DataFrame from one of the included loss models.
        # We use a single loss map for all.
        return self.ds_model.loss_map

    @_loss_map.setter
    def _loss_map(self, loss_map: pd.DataFrame) -> None:
        """
        Set the loss map.

        Parameters
        ----------
        loss_map: pd.DataFrame
            The loss map.

        """
        # Add the DataFrame to the included loss models.
        # We use a single loss map for all.
        for model in self._loss_models:
            model.loss_map = loss_map

    @property
    def _missing(self) -> set[tuple[str, str]]:
        """
        Returns the missing components.

        Returns
        -------
        set
            Set containing tuples identifying missing loss parameter
            definitions.

        """
        return self.ds_model.missing

    @_missing.setter
    def _missing(self, missing: set[tuple[str, str]]) -> None:
        """
        Assign missing parameter definitions to the loss models.

        Parameters
        ----------
        missing: set
            Set containing tuples identifying missing loss parameter
            definitions.

        """
        for model in self._loss_models:
            model.missing = missing

    def _ensure_loss_parameter_availability(self) -> None:
        """Make sure that all components have loss parameters."""
        #
        # Repair Models (currently the only type supported)
        #

        required = []
        assert self._loss_map is not None
        for dv in self.decision_variables:
            required.extend(
                [(component, dv) for component in self._loss_map['Repair']]
            )
        missing_set = set(required)

        for model in (self.ds_model, self.lf_model):
            missing_set -= model.get_available()

        if missing_set:
            self.log.warning(
                f'The loss model does not provide '
                f'loss information for the following component(s) '
                f'in the asset model: {sorted(missing_set)}.'
            )

        self._missing = missing_set


class RepairModel_Base(PelicunModel):
    """Base class for loss models."""

    __slots__ = [
        'consequence',
        'decision_variables',
        'loss_map',
        'loss_params',
        'missing',
        'sample',
    ]

    def __init__(self, assessment: AssessmentBase) -> None:
        """
        Initialize RepairModel_Base objects.

        Parameters
        ----------
        assessment: pelicun.AssessmentBase
            Parent assessment

        """
        super().__init__(assessment)

        self.loss_params: pd.DataFrame | None = None
        self.sample: pd.DataFrame | None = None
        self.consequence = 'Repair'
        self.decision_variables: tuple[str, ...] = ()
        self.loss_map: pd.DataFrame | None = None
        self.missing: set = set()

    def load_model_parameters(self, data: pd.DataFrame) -> None:
        """
        Load model parameters from a DataFrame.

        Extends those already available. Parameters already defined
        take precedence, i.e. redefinitions of parameters are ignored.

        Parameters
        ----------
        data: DataFrame
            Data with loss model information.

        """
        data.index.names = ['Loss Driver', 'Decision Variable']

        if self.loss_params is not None:
            data = pd.concat((self.loss_params, data), axis=0)

        # drop redefinitions of components
        data = (
            data.groupby(level=[0, 1]).first().transform(lambda x: x.fillna(np.nan))
        )
        # note: .groupby introduces None entries. We replace them with
        # NaN for consistency.

        self.loss_params = data

    def drop_unused_loss_parameters(self, loss_map: pd.DataFrame) -> None:
        """
        Remove loss parameter definitions.

        Applicable to component IDs not present in the loss map.

        Parameters
        ----------
        loss_map: str or pd.DataFrame or None
            Path to a csv file or DataFrame object that maps
            components IDs to their loss parameter definitions.
            Components in the asset model that are omitted from the
            provided loss map are mapped to the same ID.


        """
        if self.loss_params is None:
            return

        # <backwards compatibility>
        if 'BldgRepair' in loss_map.columns:
            loss_map['Repair'] = loss_map['BldgRepair']
            loss_map = loss_map.drop('BldgRepair', axis=1)
            self.log.warning(
                '`BldgRepair` as a loss map column name is '
                'deprecated and will be dropped in '
                'future versions of pelicun. Please use `Repair` instead.'
            )

        # get a list of unique component IDs
        cmp_set = set(loss_map['Repair'].unique())

        cmp_mask = self.loss_params.index.get_level_values(0).isin(cmp_set, level=0)
        self.loss_params = self.loss_params.iloc[cmp_mask, :]

    def remove_incomplete_components(self) -> None:
        """
        Remove incomplete components.

        Removes components that have incomplete loss model
        definitions from the loss model parameters.

        """
        if self.loss_params is None:
            return

        if ('Incomplete', '') not in self.loss_params.columns:
            return

        cmp_incomplete_idx = self.loss_params.loc[
            self.loss_params['Incomplete', ''] == 1
        ].index

        self.loss_params = self.loss_params.drop(cmp_incomplete_idx)

        if len(cmp_incomplete_idx) > 0:
            self.log.msg(
                f'\n'
                f'WARNING: Loss model information is incomplete for '
                f'the following component(s) '
                f'{cmp_incomplete_idx.to_list()}. They '
                f'were removed from the analysis.'
                f'\n',
                prepend_timestamp=False,
            )

    def get_available(self) -> set:
        """
        Get a set of components with available loss parameters.

        Returns
        -------
        set
            Set of components with available loss parameters.
        """
        if self.loss_params is not None:
            cmp_list = self.loss_params.index.to_list()
            return set(cmp_list)
        return set()

    @abstractmethod
    def convert_loss_parameter_units(self) -> None:
        """Convert previously loaded loss parameters to base units."""


class RepairModel_DS(RepairModel_Base):
    """Repair consequences for components with damage states."""

    __slots__ = ['RV_reg']

    def save_sample(
        self, filepath: str | None = None, *, save_units: bool = False
    ) -> None | pd.DataFrame | tuple[pd.DataFrame, pd.Series]:
        """
        Save or return the loss sample.

        This method handles the storage of a sample of loss estimates,
        which can either be saved directly to a file or returned as a
        DataFrame for further manipulation. When saving to a file,
        additional information such as unit conversion factors and
        column units can be included. If the data is not being saved
        to a file, the method can return the DataFrame with or without
        units as specified.

        Parameters
        ----------
        filepath: str, optional
            The path to the file where the loss sample should be
            saved. If not provided, the sample is not saved to disk
            but returned.
        save_units: bool, default: False
            Indicates whether to include a row with unit information
            in the returned DataFrame. This parameter is ignored if a
            file path is provided.

        Returns
        -------
        None or tuple
            If `filepath` is provided, the function returns None after
            saving the data.
            If no `filepath` is specified, returns:
            * DataFrame containing the loss sample.
            * Optionally, a Series containing the units for each
            column if `save_units` is True.

        """
        self.log.div()
        if filepath is not None:
            self.log.msg('Saving loss sample...')

        assert self.sample is not None
        assert self.loss_params is not None
        cmp_units = self.loss_params['DV', 'Unit'].sort_index()
        dv_units = pd.Series(
            index=self.sample.columns, name='Units', dtype='object'
        ).sort_index()

        valid_dv_types = dv_units.index.unique(level=0)
        valid_cmp_ids = dv_units.index.unique(level=1)

        for cmp_id, dv_type in cmp_units.index:
            if (dv_type in valid_dv_types) and (cmp_id in valid_cmp_ids):
                dv_units.loc[dv_type, cmp_id] = cmp_units.loc[cmp_id, dv_type]

        res = file_io.save_to_csv(
            self.sample,
            Path(filepath) if filepath is not None else None,
            units=dv_units,
            unit_conversion_factors=self._asmnt.unit_conversion_factors,
            use_simpleindex=(filepath is not None),
            log=self._asmnt.log,
        )
        if filepath is not None:
            self.log.msg('Loss sample successfully saved.', prepend_timestamp=False)
            return None

        assert isinstance(res, pd.DataFrame)

        units = res.loc['Units']
        res = res.drop('Units')
        res = res.astype(float)
        assert isinstance(res, pd.DataFrame)
        assert isinstance(units, pd.Series)

        if save_units:
            return res, units

        return res

    def load_sample(self, filepath: str | pd.DataFrame) -> dict[str, str]:
        """
        Load loss sample data.

        Parameters
        ----------
        filepath: str
          Path to an existing sample stored in a file, or dataframe
          containing the existing sample.

        Returns
        -------
        dict[str, str]
          Dictionary mapping each decision variable to its assigned
          unit.

        Raises
        ------
        ValueError
          If the columns have an invalid number of levels.

        """
        names = ['dv', 'loss', 'dmg', 'ds', 'loc', 'dir', 'uid']
        self.log.div()
        self.log.msg('Loading loss sample...')

        sample, units = file_io.load_data(
            filepath,
            self._asmnt.unit_conversion_factors,
            log=self._asmnt.log,
            return_units=True,
        )
        assert isinstance(sample, pd.DataFrame)
        assert isinstance(units, pd.Series)
        units.index.names = names
        # Obtain the DV units
        # Note: we don't need to check for consistency (all rows
        # having the same unit) since the units are extracted from a
        # single row in the CSV, affecting all subsequent rows.
        units_isolated = (
            units.reset_index()[['dv', 'Units']]
            .set_index('dv')
            .groupby('dv')['Units']
        )
        dv_units = units_isolated.first().to_dict()

        # check if `uid` level was provided
        num_levels = len(sample.columns.names)
        num_levels_without_uid = 6
        num_levels_with_uid = num_levels_without_uid + 1
        if num_levels == num_levels_without_uid:
            sample.columns.names = names[:-1]
            sample = base.dedupe_index(sample.T).T
        elif num_levels == num_levels_with_uid:
            sample.columns.names = names
        else:
            msg = (
                f'Invalid loss sample: Column MultiIndex '
                f'has an unexpected length: {num_levels}'
            )
            raise ValueError(msg)

        self.sample = sample

        self.log.msg('Loss sample successfully loaded.', prepend_timestamp=False)

        return dv_units

    def calculate(self, dmg_quantities: pd.DataFrame) -> None:  # noqa: C901
        """
        Calculate damage consequences.

        Parameters
        ----------
        dmg_quantities: DataFrame
            A table with the quantity of damage experienced in each
            damage state of each performance group at each location
            and direction. You can use the prepare_dmg_quantities
            method in the DamageModel to get such a DF.

        """
        assert self.loss_map is not None

        sample_size = len(dmg_quantities)

        # If everything is undamaged there are no losses
        if set(dmg_quantities.columns.get_level_values('ds')) == {'0'}:
            self.sample = None
            self.log.msg(
                'There is no damage---DV sample is set to None.',
                prepend_timestamp=False,
            )
            return

        # calculate the quantities for economies of scale
        self.log.msg('\nAggregating damage quantities...', prepend_timestamp=False)

        if self._asmnt.options.eco_scale['AcrossFloors']:
            if self._asmnt.options.eco_scale['AcrossDamageStates']:
                eco_levels = [0]
                eco_columns = ['cmp']

            else:
                eco_levels = [0, 4]
                eco_columns = ['cmp', 'ds']

        elif self._asmnt.options.eco_scale['AcrossDamageStates']:
            eco_levels = [0, 1]
            eco_columns = ['cmp', 'loc']

        else:
            eco_levels = [0, 1, 4]
            eco_columns = ['cmp', 'loc', 'ds']

        eco_group = dmg_quantities.groupby(level=eco_levels, axis=1)  # type: ignore
        eco_qnt = eco_group.sum().mask(eco_group.count() == 0, np.nan)
        assert eco_qnt.columns.names == eco_columns

        self.log.msg(
            'Successfully aggregated damage quantities.', prepend_timestamp=False
        )

        # apply the median functions, if needed, to get median consequences for
        # each realization
        self.log.msg(
            '\nCalculating the median repair consequences...',
            prepend_timestamp=False,
        )

        medians = self._calc_median_consequence(eco_qnt)

        self.log.msg(
            'Successfully determined median repair consequences.',
            prepend_timestamp=False,
        )

        # combine the median consequences with the samples of deviation from the
        # median to get the consequence realizations.
        self.log.msg(
            '\nConsidering deviations from the median values to obtain '
            'random DV sample...'
        )

        self.log.msg(
            'Preparing random variables for repair consequences...',
            prepend_timestamp=False,
        )
        self.RV_reg = self._create_DV_RVs(dmg_quantities.columns)  # type: ignore

        if self.RV_reg is not None:
            assert self._asmnt.options.sampling_method is not None
            self.RV_reg.generate_sample(
                sample_size=sample_size, method=self._asmnt.options.sampling_method
            )

            std_sample = base.convert_to_MultiIndex(
                pd.DataFrame(self.RV_reg.RV_sample), axis=1
            )
            std_sample.columns.names = ['dv', 'cmp', 'ds', 'loc', 'dir', 'uid']
            std_sample = std_sample.sort_index(axis=1)

        else:
            std_sample = None

        self.log.msg(
            f'\nSuccessfully generated {sample_size} realizations of '
            'deviation from the median consequences.',
            prepend_timestamp=False,
        )

        res_list = []
        key_list: list[tuple[Any, ...]] = []

        dmg_quantities.columns = dmg_quantities.columns.reorder_levels(  # type: ignore
            ['cmp', 'ds', 'loc', 'dir', 'uid']
        )
        dmg_quantities = dmg_quantities.sort_index(axis=1)

        if std_sample is not None:
            std_dvs = std_sample.columns.unique(level=0)
        else:
            std_dvs = []

        for decision_variable in self.decision_variables:  # noqa: PLR1702
            if decision_variable in std_dvs:
                assert isinstance(std_sample, pd.DataFrame)
                prob_cmp_list = std_sample[decision_variable].columns.unique(level=0)
            else:
                prob_cmp_list = []

            cmp_list: list[tuple[Any, ...]] = []

            if decision_variable not in medians:
                continue
            for component in medians[decision_variable].columns.unique(level=0):
                # check if there is damage in the component
                consequence = self.loss_map.loc[component, 'Repair']

                if component not in dmg_quantities.columns.get_level_values('cmp'):
                    continue

                ds_list = []

                for ds in (
                    medians[decision_variable]
                    .loc[:, component]
                    .columns.unique(level=0)
                ):
                    loc_list = []

                    for loc_id, loc in enumerate(
                        dmg_quantities.loc[
                            :, (component, ds)  # type: ignore
                        ].columns.unique(level=0)
                    ):
                        if (
                            self._asmnt.options.eco_scale['AcrossFloors'] is True
                        ) and (loc_id > 0):
                            break

                        if self._asmnt.options.eco_scale['AcrossFloors'] is True:
                            median_i = medians[decision_variable].loc[
                                :, (component, ds)
                            ]
                            dmg_i = dmg_quantities.loc[
                                :, (component, ds)  # type: ignore
                            ]

                            if component in prob_cmp_list:
                                assert std_sample is not None
                                std_i = std_sample.loc[
                                    :,
                                    (
                                        decision_variable,
                                        component,
                                        ds,
                                    ),  # type: ignore
                                ]
                            else:
                                std_i = None

                        else:
                            median_i = medians[decision_variable].loc[
                                :, (component, ds, loc)
                            ]
                            dmg_i = dmg_quantities.loc[
                                :, (component, ds, loc)  # type: ignore
                            ]

                            if component in prob_cmp_list:
                                assert std_sample is not None
                                std_i = std_sample.loc[
                                    :,
                                    (
                                        decision_variable,
                                        component,
                                        ds,
                                        loc,
                                    ),  # type: ignore
                                ]
                            else:
                                std_i = None

                        if std_i is not None:
                            res_list.append(dmg_i.mul(median_i, axis=0) * std_i)
                        else:
                            res_list.append(dmg_i.mul(median_i, axis=0))

                        loc_list.append(loc)

                    if self._asmnt.options.eco_scale['AcrossFloors'] is True:
                        ds_list += [
                            ds,
                        ]
                    else:
                        ds_list += [(ds, loc) for loc in loc_list]

                if self._asmnt.options.eco_scale['AcrossFloors'] is True:
                    cmp_list += [(consequence, component, ds) for ds in ds_list]
                else:
                    cmp_list += [
                        (consequence, component, ds, loc) for ds, loc in ds_list
                    ]

            if self._asmnt.options.eco_scale['AcrossFloors'] is True:
                key_list += [
                    (decision_variable, loss_cmp_i, dmg_cmp_i, ds)
                    for loss_cmp_i, dmg_cmp_i, ds in cmp_list
                ]
            else:
                key_list += [
                    (decision_variable, loss_cmp_i, dmg_cmp_i, ds, loc)
                    for loss_cmp_i, dmg_cmp_i, ds, loc in cmp_list
                ]

        lvl_names = ['dv', 'loss', 'dmg', 'ds', 'loc', 'dir', 'uid']
        dv_sample = pd.concat(res_list, axis=1, keys=key_list, names=lvl_names)

        dv_sample = dv_sample.fillna(0).convert_dtypes()

        self.log.msg('Successfully obtained DV sample.', prepend_timestamp=False)
        self.sample = dv_sample

    def convert_loss_parameter_units(self) -> None:
        """Convert previously loaded loss parameters to base units."""
        if self.loss_params is None:
            return
        units = self.loss_params['DV', 'Unit']
        arg_units = self.loss_params['Quantity', 'Unit']
        for column in self.loss_params.columns.unique(level=0):
            if not column.startswith('DS'):
                continue
            params = self.loss_params.loc[:, column].copy()
            assert isinstance(params, pd.DataFrame)
            self.loss_params.loc[:, column] = self._convert_marginal_params(
                params, units, arg_units
            ).to_numpy()

    def drop_unused_damage_states(self) -> None:
        """
        Remove unused columns.

        Remove columns from the loss model parameters corresponding
        to unused damage states.

        """
        if self.loss_params is None:
            return
        first_level = self.loss_params.columns.get_level_values(0).unique().to_list()
        ds_list = [x for x in first_level if x.startswith('DS')]
        ds_to_drop = []
        for damage_state in ds_list:
            if (
                np.all(
                    pd.isna(
                        self.loss_params.loc[
                            :,  # type: ignore
                            idx[damage_state, :],
                        ].values
                    )
                )
                # Note: When this evaluates to True, when you add `is
                # True` on the right it suddenly evaluates to
                # False. We need to figure out why this is happening,
                # but the way it's written now does what we want in
                # each case.
            ):
                ds_to_drop.append(damage_state)  # noqa: PERF401

        self.loss_params = self.loss_params.drop(columns=ds_to_drop, level=0)

    def _create_DV_RVs(  # noqa: N802, C901
        self, cases: pd.MultiIndex
    ) -> uq.RandomVariableRegistry | None:
        """
        Prepare the random variables.

        Prepare the random variables associated with decision
        variables, such as repair cost and time.

        Parameters
        ----------
        cases: MultiIndex
            Index with cmp-loc-uid-dir-ds descriptions that identify
            the RVs we need for the simulation.

        Returns
        -------
        RandomVariableRegistry or None
            A RandomVariableRegistry containing all the generated
            random variables necessary for the simulation. If no
            random variables are generated (due to missing parameters
            or conditions), returns None.

        """
        # Convert the MultiIndex to a DataFrame
        case_df = pd.DataFrame(index=cases).reset_index()
        # maps `cmp` to array of damage states
        damage_states = case_df.groupby('cmp')['ds'].unique().to_dict()
        # maps `cmp`-`ds` to tuple of `loc`-`dir`-`uid` tuples
        loc_dir_uids = (
            case_df.groupby(['cmp', 'ds'])[['loc', 'dir', 'uid']]
            .apply(
                lambda x: tuple(  # type: ignore
                    zip(
                        x['loc'],
                        x['dir'],
                        x['uid'],
                    )
                )
            )
            .to_dict()
        )
        damaged_components = set(cases.get_level_values('cmp'))

        rv_reg = uq.RandomVariableRegistry(self._asmnt.options.rng)

        rv_count = 0

        # for each component in the loss map
        assert isinstance(self.loss_map, pd.DataFrame)
        for component, consequence in self.loss_map['Repair'].items():
            # if that component does not have realized damage states,
            # skip it (e.g., this can happen when there is only
            # `collapse`).
            if component not in damaged_components:
                continue

            # for each DV
            for decision_variable in self.decision_variables:
                # If loss parameters are missing for that consequence,
                # don't estimate losses for it. A warning has already
                # been issued for what is missing.
                if (consequence, decision_variable) in self.missing:
                    continue

                # If loss parameters are missing for that consequence,
                # for this particular loss model, they will exist in
                # the other(s).
                assert self.loss_params is not None
                if (consequence, decision_variable) not in self.loss_params.index:
                    continue

                # load the corresponding parameters
                parameters = (
                    self.loss_params.loc[(consequence, decision_variable), :]
                    .dropna()
                    .to_dict()
                )

                for ds in damage_states[component]:
                    if ds == '0':
                        continue

                    ds_family = parameters.get((f'DS{ds}', 'Family'))
                    ds_theta = [
                        theta
                        for t_i in range(3)
                        if (theta := parameters.get((f'DS{ds}', f'Theta_{t_i}')))
                        is not None
                    ]
                    # If there is no RV family we don't need an RV
                    if ds_family is None:
                        continue

                    # If the first parameter is controlled by a function, we use
                    # 1.0 in its place and will scale the results in a later
                    # step
                    if isinstance(ds_theta[0], str) and '|' in ds_theta[0]:
                        ds_theta[0] = 1.0

                    loc_dir_uid = loc_dir_uids[component, ds]

                    for loc, direction, uid in loc_dir_uid:
                        # assign RVs
                        rv_reg.add_RV(
                            uq.rv_class_map(ds_family)(  # type: ignore
                                name=(
                                    f'{decision_variable}-{component}-'
                                    f'{ds}-{loc}-{direction}-{uid}'
                                ),
                                theta=np.array(ds_theta),
                                truncation_limits=np.array([0.0, np.nan]),
                            )
                        )
                        rv_count += 1

        # assign Time-Cost correlation whenever applicable
        rho = self._asmnt.options.rho_cost_time
        if rho != 0.0:
            for rv_tag in rv_reg.RV:
                if not rv_tag.startswith('Cost'):
                    continue
                component = rv_tag.split('-')[1]
                ds = rv_tag.split('-')[2]
                loc = rv_tag.split('-')[3]
                direction = rv_tag.split('-')[4]
                uid = rv_tag.split('-')[5]
                time_rv_tag = rv_tag.replace('Cost', 'Time')
                if time_rv_tag in rv_reg.RV:
                    rv_reg.add_RV_set(
                        uq.RandomVariableSet(
                            f'DV-{component}-{ds}-{loc}-{direction}-{uid}_set',
                            list(rv_reg.RVs([rv_tag, time_rv_tag]).values()),
                            np.array([[1.0, rho], [rho, 1.0]]),
                        )
                    )

        self.log.msg(
            f'\n{rv_count} random variables created.', prepend_timestamp=False
        )

        if rv_count > 0:
            return rv_reg
        return None

    def _calc_median_consequence(self, eco_qnt: pd.DataFrame) -> dict:  # noqa: C901
        """
        Calculate median reiapr consequences.

        Calculate the median repair consequences for each loss
        component based on its quantity realizations and the
        associated loss parameters.

        Parameters
        ----------
        eco_qnt: DataFrame
            A DataFrame containing quantity realizations for various
            components and damage states, appropriately grouped to
            account for economies of scale.

        decision_variables: list
            Defines the decision variables to be included in the loss
            calculations. Defaults to those supported, but fewer can be
            used if desired. When fewer are used, the loss parameters for
            those not used will not be required.

        Returns
        -------
        dict
            A dictionary where keys are the types of decision
            variables (DV) like 'COST' or 'TIME', and values are
            DataFrames containing the median consequences for each
            component and damage state. These DataFrames are structured
            with MultiIndex columns that may include 'cmp' (component),
            'ds' (damage state), and potentially 'loc' (location),
            depending on the way economies of scale are handled.

        Raises
        ------
        ValueError
            If any loss driver types or distribution types are not
            recognized, or if the parameters are incomplete or
            unsupported.

        """
        medians = {}

        for decision_variable in self.decision_variables:
            cmp_list = []
            median_list = []

            assert self.loss_map is not None
            for loss_cmp_id, loss_cmp_name in self.loss_map['Repair'].items():
                if (loss_cmp_name, decision_variable) in self.missing:
                    continue

                if loss_cmp_id not in eco_qnt.columns.get_level_values(0).unique():
                    continue

                ds_list = []
                sub_medians = []

                assert self.loss_params is not None
                for ds in self.loss_params.columns.get_level_values(0).unique():
                    if not ds.startswith('DS'):
                        continue

                    ds_id = ds[2:]

                    if ds_id == '0':
                        continue

                    loss_params_ds = self.loss_params.loc[
                        (loss_cmp_name, decision_variable), ds
                    ]

                    # check if theta_0 is defined
                    theta_0 = loss_params_ds.get('Theta_0', np.nan)

                    if pd.isna(theta_0):
                        continue

                    # check if the distribution type is supported
                    family = loss_params_ds.get('Family', np.nan)

                    if (not pd.isna(family)) and (
                        family not in {'normal', 'lognormal', 'deterministic'}
                    ):
                        msg = (
                            f'Loss Distribution of type {family} ' f'not supported.'
                        )
                        raise ValueError(msg)

                    # If theta_0 is a scalar
                    try:
                        theta_0 = float(theta_0)

                        if pd.isna(loss_params_ds.get('Family', np.nan)):
                            # if theta_0 is constant, then use it directly
                            f_median = _prep_constant_median_DV(theta_0)

                        else:
                            # otherwise use a constant 1.0 as the median
                            # The random variable will be generated as a
                            # variation from this 1.0 and added in a later step.
                            f_median = _prep_constant_median_DV(1.0)

                    except ValueError:
                        # otherwise, use the multilinear function
                        all_vals = np.array(
                            [val.split(',') for val in theta_0.split('|')],
                            dtype=float,
                        )
                        medns = all_vals[0]
                        qnts = all_vals[1]
                        f_median = _prep_bounded_multilinear_median_DV(medns, qnts)

                    # get the corresponding aggregate damage quantities
                    # to consider economies of scale
                    if 'ds' in eco_qnt.columns.names:
                        avail_ds = eco_qnt.loc[
                            :,  # type: ignore
                            loss_cmp_id,
                        ].columns.unique(level=0)

                        if ds_id not in avail_ds:
                            continue

                        eco_qnt_i = eco_qnt.loc[
                            :,  # type: ignore
                            (loss_cmp_id, ds_id),
                        ].copy()

                    else:
                        eco_qnt_i = eco_qnt.loc[
                            :,  # type: ignore
                            loss_cmp_id,
                        ].copy()

                    if isinstance(eco_qnt_i, pd.Series):
                        eco_qnt_i = eco_qnt_i.to_frame()
                        eco_qnt_i.columns = ['X']
                        eco_qnt_i.columns.name = 'del'

                    # generate the median values for each realization
                    eco_qnt_i.loc[:, :] = f_median(eco_qnt_i.values)  # type: ignore

                    sub_medians.append(eco_qnt_i)
                    ds_list.append(ds_id)

                if len(ds_list) > 0:
                    # combine medians across damage states into one DF
                    median_list.append(pd.concat(sub_medians, axis=1, keys=ds_list))
                    cmp_list.append(loss_cmp_id)

            if len(cmp_list) > 0:
                # combine medians across components into one DF
                result = pd.concat(median_list, axis=1, keys=cmp_list)

                # remove the extra column header level
                if 'del' in result.columns.names:
                    result.columns = result.columns.droplevel('del')

                # name the remaining column header levels
                if eco_qnt.columns.names == ['cmp', 'ds', 'loc']:
                    result.columns.names = ['cmp', 'ds', 'loc']
                else:
                    result.columns.names = ['cmp', 'ds']

                # save the results to the returned dictionary
                medians.update({decision_variable: result})

        return medians


class RepairModel_LF(RepairModel_Base):
    """Repair consequences for components with loss functions."""

    __slots__ = []

    def calculate(
        self,
        demand_sample: pd.DataFrame,
        cmp_sample: dict,
        cmp_marginal_params: pd.DataFrame,
        demand_offset: dict,
        nondirectional_multipliers: dict,
    ) -> None:
        """
        Calculate repair consequences.

        Parameters
        ----------
        demand_sample: pd.DataFrame
            The sample of the demand model to be used for the inputs
            of the loss functions.
        cmp_sample: dict
            Dict mapping each `cmp`-`loc`-`dir`-`uid` to the component
            quantity realizations in the asset model in the form of
            pd.Series objects.
        cmp_marginal_params: pd.DataFrame
            Dataframe containing component marginal distribution
            parameters.
        demand_offset: dict
            Dictionary specifying the demand offset.
        nondirectional_multipliers: dict
            Dictionary specifying the non directional multipliers used
            to combine the directional demands.


        """
        if self.loss_params is None:
            return

        assert self.loss_map is not None
        loss_map = self.loss_map['Repair'].to_dict()
        sample_size = len(demand_sample)

        # TODO(JVM): this can be taken out and simply passed as blocks in
        # the arguments, and cast to a dict in here. Index can be
        # obtained from there.
        index = [
            x
            for x in cmp_marginal_params.index.get_level_values(0)
            if loss_map.get(x) in self.loss_params.index
        ]
        # If `Blocks` information is unspecified add one block per
        # component.
        if 'Blocks' not in cmp_marginal_params.columns:
            cmp_marginal_params['Blocks'] = 1
        blocks = cmp_marginal_params.loc[index, 'Blocks'].to_dict()

        performance_group_dict = {}
        for (component, location, direction, uid), num_blocks in blocks.items():
            for decision_variable in self.decision_variables:
                if (component, decision_variable) in self.missing:
                    continue
                performance_group_dict[
                    (component, decision_variable), location, direction, uid
                ] = num_blocks

        if not performance_group_dict:
            self.log.msg(
                'No loss function-driven components---LF sample is set to None.',
                prepend_timestamp=False,
            )
            return

        performance_group = pd.DataFrame(  # type: ignore
            performance_group_dict.values(),
            index=performance_group_dict.keys(),
            columns=['Blocks'],
        )
        performance_group.index.names = ['cmp', 'loc', 'dir', 'uid']

        required_edps = base.invert_mapping(
            _get_required_demand_type(
                self.loss_params, performance_group, demand_offset
            )
        )

        available_edps = (
            pd.DataFrame(index=demand_sample.columns)
            .reset_index()
            .groupby(['type', 'loc'])['dir']
            .agg(lambda x: list(set(x)))
            .to_dict()
        )

        # Raise an error if demand sample is missing necessary entries.
        _verify_edps_available(available_edps, set(required_edps.values()))

        demand_dict = _assemble_required_demand_data(
            set(required_edps.values()),
            nondirectional_multipliers,
            demand_sample,
        )

        self.log.msg(
            '\nCalculating the median repair consequences...',
            prepend_timestamp=False,
        )

        medians = self._calc_median_consequence(
            performance_group, loss_map, required_edps, demand_dict, cmp_sample
        )

        self.log.msg(
            'Successfully determined median repair consequences.',
            prepend_timestamp=False,
        )

        self.log.msg(
            '\nConsidering deviations from the median values to obtain '
            'random DV sample...'
        )

        self.log.msg(
            'Preparing random variables for repair cost and time...',
            prepend_timestamp=False,
        )

        rv_reg = self._create_DV_RVs(medians.columns)  # type: ignore
        if rv_reg is not None:
            assert self._asmnt.options.sampling_method is not None
            rv_reg.generate_sample(
                sample_size=sample_size, method=self._asmnt.options.sampling_method
            )

            std_sample = base.convert_to_MultiIndex(
                pd.DataFrame(rv_reg.RV_sample), axis=1
            )
            std_sample.columns.names = [
                'dv',
                'loss',
                'dmg',
                'loc',
                'dir',
                'uid',
                'block',
            ]
            std_sample = std_sample.sort_index(axis=1)
            sample = (medians * std_sample).combine_first(medians)

        else:
            sample = medians

        self.log.msg(
            f'\nSuccessfully generated {sample_size} realizations of '
            'deviation from the median consequences.',
            prepend_timestamp=False,
        )

        # sum up the block losses
        sample = sample.groupby(  # type: ignore
            by=['dv', 'loss', 'dmg', 'loc', 'dir', 'uid'], axis=1
        ).sum()

        self.log.msg('Successfully obtained DV sample.', prepend_timestamp=False)
        self.sample = sample

        return

    def convert_loss_parameter_units(self) -> None:
        """Convert previously loaded loss parameters to base units."""
        if self.loss_params is None:
            return
        units = self.loss_params['DV', 'Unit']
        arg_units = self.loss_params['Demand', 'Unit']
        column = 'LossFunction'
        params = self.loss_params[column].copy()
        assert isinstance(params, pd.DataFrame)
        self.loss_params.loc[:, column] = self._convert_marginal_params(
            params,
            units,
            arg_units,
            divide_units=False,
        ).to_numpy()
        return

    def _calc_median_consequence(
        self,
        performance_group: pd.DataFrame,
        loss_map: dict,
        required_edps: dict,
        demand_dict: dict,
        cmp_sample: dict,
    ) -> pd.DataFrame:
        """
        Calculate median repair consequences.

        Calculates the median repair consequences for each loss
        function-driven component based on its quantity realizations
        and the associated loss parameters.

        Parameters
        ----------
        performance_group: pd.DataFrame
            Dataframe with a single column `Blocks` containing an
            integer for the number of blocks of each
            (`cmp`-`decision_variable`)-`loc`-`dir`-`uid`.
        loss_map: dict
            Dictionary mpping component IDs, `cmp`, to their repair
            consequences.
        required_edps: dict
            Dictionary mapping (`cmp`-`realization`)-`loc`-`dir`-`uid`
            (each entry of the `performance_group`'s index) with the
            EDP name (e.g., `PFA-1-1`) that should be used as its loss
            function input.
        demand_dict: dict
            Dictionary mapping each EDP name to the values in the
            demand sample in the form of numpy arrays.
        cmp_sample: dict
            Dict mapping each `cmp`-`loc`-`dir`-`uid` to the component
            quantity realizations in the asset model in the form of
            pd.Series objects.


        Returns
        -------
        pd.DataFrame
            Dataframe with medial loss for loss-function driven
            components.

        Raises
        ------
        ValueError
            If loss function interpolation fails.

        """
        medians_dict = {}

        # for each component in the asset model
        component: str
        decision_variable: str
        location: str
        direction: str
        uid: str
        blocks: int
        for (  # type: ignore
            (component, decision_variable),
            location,
            direction,
            uid,
        ), blocks in performance_group['Blocks'].items():
            consequence = loss_map[component]
            edp = required_edps[
                (consequence, decision_variable), location, direction, uid
            ]
            edp_values = demand_dict[edp]
            assert self.loss_params is not None
            loss_function_str = self.loss_params.loc[
                (component, decision_variable), ('LossFunction', 'Theta_0')
            ]
            assert isinstance(loss_function_str, str)
            try:
                median_loss = base.stringterpolation(loss_function_str)(edp_values)
            except ValueError as exc:
                msg = (
                    f'Loss function interpolation for consequence '
                    f'`{consequence}-{decision_variable}` has failed. '
                    f'Ensure a sufficient interpolation domain  '
                    f'for the X values (those after the `|` symbol)  '
                    f'and verify the X-value and Y-value lengths match.'
                )
                raise ValueError(msg) from exc
            for block in range(blocks):
                medians_dict[
                    decision_variable,
                    consequence,
                    component,
                    location,
                    direction,
                    uid,
                    str(block),
                ] = (
                    median_loss
                    * cmp_sample[component, location, direction, uid].to_numpy()
                    / float(blocks)
                )

        medians = pd.DataFrame(medians_dict)
        medians.columns.names = ['dv', 'loss', 'dmg', 'loc', 'dir', 'uid', 'block']
        return medians.sort_index(axis=1)

    def _create_DV_RVs(  # noqa: N802
        self, cases: pd.MultiIndex
    ) -> uq.RandomVariableRegistry | None:
        """
        Prepare the decision variable random variables.

        Prepares the random variables associated with decision
        variables, such as repair cost and time.

        Parameters
        ----------
        cases: MultiIndex
            Index with `dv`-`loss`-`dmg`-`loc`-`dir`-`uid`
            entries that identify the RVs we need for the
            simulation (columns of the `medians` dataframe).

        Returns
        -------
        RandomVariableRegistry or None
            A RandomVariableRegistry containing all the generated
            random variables necessary for the simulation. If no
            random variables are generated (due to missing parameters
            or conditions), returns None.

        """
        rv_reg = uq.RandomVariableRegistry(self._asmnt.options.rng)

        rv_count = 0

        # for each component in the loss map
        for (
            decision_variable,
            consequence,
            component,
            location,
            direction,
            uid,
            block,
        ) in cases:
            # load the corresponding parameters
            assert self.loss_params is not None
            parameters = self.loss_params.loc[(consequence, decision_variable), :]

            if ('LossFunction', 'Family') not in parameters:
                # Everything is deterministic, no need to create RVs.
                continue
            family = parameters.loc['LossFunction', 'Family']
            theta = [
                theta_value
                for t_i in range(3)
                if (theta_value := parameters.get(('LossFunction', f'Theta_{t_i}')))
                is not None
            ]
            # If there is no RV family we don't need an RV
            if pd.isna(family):
                continue

            # Since the first parameter is controlled by a function,
            # we use 1.0 in its place and will scale the results in a
            # later step.
            theta[0] = 1.0  # type: ignore

            # assign RVs
            rv_reg.add_RV(
                uq.rv_class_map(family)(  # type: ignore
                    name=(
                        f'{decision_variable}-{consequence}-'
                        f'{component}-{location}-{direction}-{uid}-{block}'
                    ),
                    theta=np.array(theta),
                    truncation_limits=np.array([0.0, np.nan]),
                )
            )
            rv_count += 1

        # assign Time-Cost correlation whenever applicable
        rho = self._asmnt.options.rho_cost_time
        if rho != 0.0:
            for rv_tag in rv_reg.RV:
                if not rv_tag.startswith('Cost'):
                    continue
                split = rv_tag.split('-')
                consequence = split[1]
                component = split[2]
                location = split[3]
                direction = split[4]
                uid = split[5]
                block = split[6]
                time_rv_tag = rv_tag.replace('Cost', 'Time')
                if time_rv_tag in rv_reg.RV:
                    rv_reg.add_RV_set(
                        uq.RandomVariableSet(
                            (
                                f'DV-{consequence}-{component}-'
                                f'{location}-{direction}-{uid}-{block}_set'
                            ),
                            list(rv_reg.RVs([rv_tag, time_rv_tag]).values()),
                            np.array([[1.0, rho], [rho, 1.0]]),
                        )
                    )

        self.log.msg(
            f'\n{rv_count} random variables created.', prepend_timestamp=False
        )

        if rv_count > 0:
            return rv_reg
        return None


def _prep_constant_median_DV(median: float) -> Callable:  # noqa: N802
    """
    Return a constant median Decision Variable (DV) function.

    Parameters
    ----------
    median: float
        The median DV for a consequence function with fixed median.

    Returns
    -------
    callable
        A function that returns the constant median DV for all component
        quantities.

    """

    def f(*args):  # noqa: ANN002, ANN202, ARG001
        return median

    return f


def _prep_bounded_multilinear_median_DV(  # noqa: N802
    medians: np.ndarray, quantities: np.ndarray
) -> Callable:
    """
    Return a bounded multilinear median Decision Variable (DV) function.

    The median DV equals the min and max values when the quantity is
    outside of the prescribed quantity bounds. When the quantity is within the
    bounds, the returned median is calculated by linear interpolation.

    Parameters
    ----------
    medians: ndarray
        Series of values that define the y coordinates of the multilinear DV
        function.
    quantities: ndarray
        Series of values that define the component quantities corresponding to
        the series of medians and serving as the x coordinates of the
        multilinear DV function.

    Returns
    -------
    callable
        A function that returns the median DV given the quantity of damaged
        components.

    """

    def f(quantity):  # noqa: ANN001, ANN202
        if quantity is None:
            msg = (
                'A bounded linear median Decision Variable function called '
                'without specifying the quantity of damaged components'
            )
            raise ValueError(msg)

        q_array = np.asarray(quantity, dtype=np.float64)

        # calculate the median consequence given the quantity of damaged
        # components
        return np.interp(q_array, quantities, medians)

    return f


def _is_for_lf_model(data: pd.DataFrame) -> bool:
    """
    Determine if the data are for the lf_model.

    Parameters
    ----------
    data: pd.DataFrame
        Data to be checked.

    Returns
    -------
    bool
        Whether the data are for the lf_model.

    """
    return 'LossFunction' in data.columns.get_level_values(0)


def _is_for_ds_model(data: pd.DataFrame) -> bool:
    """
    Determine if the data are for the ds_model.

    Parameters
    ----------
    data: pd.DataFrame
        Data to be checked.

    Returns
    -------
    bool
        Whether the data are for the ds_model.

    """
    return 'DS1' in data.columns.get_level_values(0)



# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay

"""Temporary solution that provides regional simulation capability to Pelicun."""

from __future__ import annotations

import contextlib
import json
import os
import tempfile
import time
from pathlib import Path

import joblib
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm import tqdm

from pelicun.assessment import Assessment
from pelicun.auto import auto_populate
from pelicun.file_io import substitute_default_path
from pelicun.tools.NNR import NNR


def unique_list(x: pd.Series) -> str:
    """
    Return unique values in a pandas Series as a comma-separated string.

    Parameters
    ----------
    x : pd.Series
        pandas Series containing values to extract unique elements from

    Returns
    -------
    str
        Comma-separated string of unique values, or single value if only one unique value exists

    """
    vals = x.unique().astype(str)

    if len(vals) > 1:
        return f'{",".join(vals)}'

    return f'{vals[0]}'


def format_elapsed_time(start_time: float) -> str:
    """
    Format elapsed time from a start timestamp to current time as hh:mm:ss.

    Parameters
    ----------
    start_time : float
        Start time as a float timestamp (from time.time())

    Returns
    -------
    str
        Formatted elapsed time string in hh:mm:ss format

    """
    elapsed = time.time() - start_time
    hours = int(elapsed // 3600)
    minutes = int((elapsed % 3600) // 60)
    seconds = int(elapsed % 60)
    return f'{hours:02d}:{minutes:02d}:{seconds:02d}'


@contextlib.contextmanager
def tqdm_joblib(tqdm_object: tqdm) -> contextlib.Generator[None, None, None]:
    """
    Context manager to patch joblib to report progress into a tqdm progress bar.

    This function temporarily replaces joblib's BatchCompletionCallBack to update
    the provided tqdm progress bar with batch completion information during
    parallel processing.

    Parameters
    ----------
    tqdm_object : tqdm
        tqdm progress bar object to update with progress information

    Yields
    ------
    None
        Context manager yields control to the calling code

    """

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):  # noqa: ANN204, ANN002, ANN003
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


def process_buildings_chunk(
    bldg_df_chunk: pd.DataFrame,
    grid_points: pd.DataFrame,
    grid_data: pd.DataFrame,
    sample_size_demand: int,
    sample_size_damage: int,
    dl_method: str,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Process a chunk of buildings through the complete regional simulation pipeline.

    This function performs event-to-building mapping, building-to-archetype mapping,
    damage calculation, and loss calculation for a subset of buildings. It uses
    nearest neighbor regression to map grid-based intensity measures to building
    locations, then applies Pelicun assessment methods to calculate damage and losses.

    Parameters
    ----------
    bldg_df_chunk : pd.DataFrame
        DataFrame containing building inventory data for this chunk,
        must include Longitude, Latitude, and building characteristics
    grid_points : pd.DataFrame
        DataFrame with grid point coordinates (Longitude, Latitude)
    grid_data : pd.DataFrame
        DataFrame with intensity measure data for each grid point
    sample_size_demand : int
        Number of demand realizations available
    sample_size_damage : int
        Number of damage realizations to generate
    dl_method : str
        Damage and loss methodology

    Returns
    -------
    tuple
        demand_sample : pd.DataFrame
            DataFrame with demand sample results (intensity measures)
        damage_df : pd.DataFrame
            DataFrame with damage state results for each component
        repair_costs : pd.DataFrame
            DataFrame with repair cost estimates
        repair_times : pd.DataFrame
            DataFrame with repair time estimates

    """
    # 3 Event-to-Building Mapping
    # Map event IMs to building centroids using the nearest neighbor method

    X = grid_points[['Longitude', 'Latitude']].to_numpy()  # noqa: N806
    Z = grid_data.T.to_numpy()  # noqa: N806

    X_t = bldg_df_chunk[['Longitude', 'Latitude']].to_numpy()  # noqa: N806

    Z_hat = NNR(X_t, X, Z, sample_size=-1, n_neighbors=8, weight='distance2')  # noqa: N806

    # Prepare IM information as demands in the Pelicun format

    # TODO: allow for more flexible columns  # noqa: TD002
    demand_sample = pd.DataFrame(
        np.vstack([Z_hat.T, np.full((1, Z_hat.shape[0]), 'g')]),
        columns=[f'PGA-{loc}-1' for loc in bldg_df_chunk.index],
        index=[*list(range(sample_size_demand)), 'Units'],
    )

    # 4 Building-to-Archetype Mapping
    # Use Pelicun and the built-in mapping for Hazus IM-based damage models to map each building in the inventory to an archetype

    auto_script_path = substitute_default_path(
        [f'PelicunDefault/{dl_method}/pelicun_config.py']
    )[0]

    CMP_list = []  # noqa: N806

    for bldg_id, row_data in bldg_df_chunk.iloc[:].iterrows():
        _, CMP = auto_populate(  # noqa: N806
            {
                'GeneralInformation': dict(row_data),
                'assetType': 'Buildings',
                'Applications': {
                    'DL': {
                        'ApplicationData': {
                            'coupled_EDP': True,
                            'lifeline_facility': True,
                            'ground_failure': False,
                        }
                    }
                },
            },
            auto_script_path,
        )

        CMP['Location'] = bldg_id

        CMP_list.append(CMP)

    cmp_marginals = pd.concat(CMP_list)

    cmp_marginals_raw = cmp_marginals.copy()

    cmp_marginals = cmp_marginals_raw.groupby(cmp_marginals_raw.index).agg(
        {
            'Units': unique_list,
            'Location': unique_list,
            'Direction': unique_list,
            'Theta_0': unique_list,
        }
    )

    # 5 Calculate Damage
    # Calculate damage with a deterministic inventory realization

    # initialize assessment object
    PAL = Assessment(  # noqa: N806
        {
            'LogFile': 'pelicun_log.txt',
            'Verbose': True,
            'NonDirectionalMultipliers': {'ALL': 1.0},
        }
    )

    # load demands
    PAL.demand.load_sample(demand_sample)

    PAL.demand.calibrate_model({'ALL': {'DistributionFamily': 'empirical'}})

    PAL.demand.generate_sample(
        {'SampleSize': sample_size_damage, 'PreserveRawOrder': True}
    )

    # load component assignment
    PAL.asset.load_cmp_model({'marginals': cmp_marginals})

    PAL.asset.generate_cmp_sample()

    # get the path to the built-in fragility functions
    component_db_path = substitute_default_path(
        [f'PelicunDefault/{dl_method}/fragility.csv']
    )[0]

    # import the required fragility functions
    cmp_set = PAL.asset.list_unique_component_ids()

    PAL.damage.load_model_parameters(
        [
            component_db_path,
        ],
        cmp_set,
    )

    # run the damage calculation
    PAL.damage.calculate()

    # retrieve damage information
    damage_sample, damage_units = PAL.damage.save_sample(save_units=True)

    damage_units = damage_units.to_frame().T

    # aggregate across uid
    # this is trivial since we don't have multiple identical components at the same location
    damage_units = damage_units.groupby(
        level=['cmp', 'loc', 'dir', 'ds'], axis=1
    ).first()

    damage_groupby_uid = damage_sample.groupby(
        level=['cmp', 'loc', 'dir', 'ds'], axis=1
    )
    damage_sample = damage_groupby_uid.sum().mask(
        damage_groupby_uid.count() == 0, np.nan
    )

    # aggregate across dir
    # also trivial, all results are in dir 1

    damage_groupby = damage_sample.groupby(level=['cmp', 'loc', 'ds'], axis=1)

    damage_units = damage_units.groupby(level=['cmp', 'loc', 'ds'], axis=1).first()
    grp_damage = damage_groupby.sum().mask(damage_groupby.count() == 0, np.nan)

    # replace non-zero values with 1
    # this is probably not making any meaningful changes since we have 1 ea quantity of each component
    # Honestly, I am not quite sure why we need this step...
    grp_damage = grp_damage.mask(grp_damage.astype(np.float64).to_numpy() > 0, 1)

    # get the corresponding DS for each column
    ds_list = grp_damage.columns.get_level_values('ds').astype(int)

    # replace ones with the corresponding DS in each cell
    grp_damage = grp_damage.mul(ds_list, axis=1)

    # aggregate across damage state indices
    damage_groupby_2 = grp_damage.groupby(level=['cmp', 'loc'], axis=1)

    # choose the max value
    # i.e., the governing DS for each comp-loc pair
    # Note that in each realization, each component will have only one non-zero damage state result,
    # so this is not picking the max damage state, but rather picking the realized damage state
    grp_damage = damage_groupby_2.max().mask(damage_groupby_2.count() == 0, np.nan)

    # aggregate units to the same format
    # assume identical units across locations for each comp
    damage_units = damage_units.groupby(level=['cmp', 'loc'], axis=1).first()

    grp_damage = grp_damage.astype(int).T.reorder_levels(['loc', 'cmp'])

    damage_df = grp_damage.copy()

    # assuming there's only one component in each building, we can drop the component info
    damage_df = damage_df.groupby(level='loc').sum()
    damage_df.index = damage_df.index.astype(int)
    damage_df = damage_df.sort_index()

    # 6 Calculate Losses

    # get the path to the built-in consequence functions
    consequence_db_path = substitute_default_path(
        [f'PelicunDefault/{dl_method}/consequence_repair.csv']
    )[0]

    # Hazus consequence functions depend only on occupancy class
    # we need to list building IDs for each occupancy type first
    loss_groups = bldg_df_chunk[['StructureType', 'OccupancyClass']].copy()
    loss_groups['IDs'] = loss_groups.index

    loss_groups = loss_groups.groupby(['OccupancyClass']).agg({'IDs': unique_list})

    # create a lookup table to find which fragility IDs are at which location
    dmg_sample = PAL.damage.save_sample()
    cmp_loc = dmg_sample.groupby(level=['cmp', 'loc'], axis=1).first()

    cmp_lookup = pd.Series(
        cmp_loc.columns.get_level_values('cmp'),
        index=cmp_loc.columns.get_level_values('loc').astype(int),
    ).sort_index()

    # Loss calculation is a bit complicated now. I need to add a new feature to Pelicun to make it work as smoothly as the damage does.
    # What I do below is brute force, but it works, and still takes only a few minutes to run.
    # I'll enhance Pelicun in the coming weeks to do more sophisticated loss mapping and be able to handle the following calculations faster without the for loop

    # we'll collect the results in this dict
    loss_results = {'Cost': [], 'Time': []}

    # start by extracting the full damage sample and preserving it
    full_dmg_sample = PAL.damage.save_sample()

    for occ_type, raw_building_ids in loss_groups.iloc[:].iterrows():
        # convert the building id list to a numpy array of ints
        building_ids = np.array(raw_building_ids.to_numpy()[0].split(',')).astype(
            int
        )

        # and make sure those IDs are in the component lookup table
        building_ids = [
            building_id
            for building_id in building_ids
            if building_id in cmp_lookup.index
        ]

        # load only the subset of the damage sample that is needed for the calculation
        idx = pd.IndexSlice
        dmg_subset = full_dmg_sample.loc[
            :, idx[:, np.array(building_ids, dtype=str), :, :, :]
        ]
        PAL.damage.load_sample(dmg_subset)

        # create a loss map that maps every fragility ID to a consequence for the given occupancy type
        loss_cmp = f'LF.{occ_type}'

        drivers = []
        loss_models = []

        dmg_cmps = dmg_subset.columns.get_level_values('cmp').unique()

        for cmp_id in dmg_cmps.unique():
            drivers.append(f'{cmp_id}')
            loss_models.append(loss_cmp)

        loss_map = pd.DataFrame(loss_models, columns=['Repair'], index=drivers)

        decision_variables = ['Cost', 'Time']

        PAL.loss.decision_variables = decision_variables
        PAL.loss.add_loss_map(loss_map, loss_map_policy=None)
        PAL.loss.load_model_parameters(
            [
                consequence_db_path,
            ]
        )

        PAL.loss.calculate()

        # - - - - - -

        repair_sample, repair_units = PAL.loss.save_sample(save_units=True)

        # aggregate across uid
        # this is trivial since we don't have multiple identical components at the same location
        repair_units = repair_units.groupby(
            level=['dv', 'loss', 'dmg', 'ds', 'loc', 'dir']
        ).first()

        repair_groupby_uid = repair_sample.groupby(
            level=['dv', 'loss', 'dmg', 'ds', 'loc', 'dir'], axis=1
        )

        repair_sample = repair_groupby_uid.sum().mask(
            repair_groupby_uid.count() == 0, np.nan
        )

        # now aggregate across loss, dmg, damage state, and direction
        # all of those are only one value per location, so they do not provide additional information
        repair_groupby = repair_sample.groupby(level=['dv', 'loc'], axis=1)

        repair_units = repair_units.groupby(level=['dv', 'loc']).first()

        grp_repair = repair_groupby.sum().mask(repair_groupby.count() == 0, np.nan)

        # - - - -

        # append the results to the main dict
        for DV_type in ['Cost', 'Time']:  # noqa: N806
            grp_repair_dv = grp_repair[DV_type].copy()
            grp_repair_dv.columns = grp_repair_dv.columns.astype(int)

            # we did not preserve the units for now; we can add them here if needed

            loss_results[DV_type].append(grp_repair_dv.loc[:, building_ids])

    repair_costs = pd.concat(loss_results['Cost'], axis=1).sort_index(axis=1).T
    repair_times = pd.concat(loss_results['Time'], axis=1).sort_index(axis=1).T

    # finish by putting the full damage sample back to the assessment object
    PAL.damage.load_sample(full_dmg_sample)

    return demand_sample, damage_df, repair_costs, repair_times


def process_and_save_chunk(
    i: int,
    chunk: pd.DataFrame,
    temp_dir: str,
    grid_points: pd.DataFrame,
    grid_data: pd.DataFrame,
    sample_size_demand: int,
    sample_size_damage: int,
    dl_method: str,
) -> None:
    """
    Process a single chunk of buildings and save results to temporary compressed CSV files.

    This function serves as a wrapper around process_buildings_chunk that handles
    the file I/O operations for parallel processing. It processes a chunk of buildings
    through the complete simulation pipeline and saves the results to temporary files
    for later aggregation.

    Parameters
    ----------
    i : int
        Chunk index number used for naming output files
    chunk : pd.DataFrame
        DataFrame containing building inventory data for this specific chunk
    temp_dir : str
        Path to temporary directory where results will be saved
    grid_points : pd.DataFrame
        DataFrame with grid point coordinates (Longitude, Latitude)
    grid_data : pd.DataFrame
        DataFrame with intensity measure data for each grid point
    sample_size_demand : int
        Number of demand realizations available
    sample_size_damage : int
        Number of damage realizations to generate
    dl_method : str
        Damage and loss methodology

    """
    # Process buildings through steps 3-6
    demand_sample_chunk, damage_df_chunk, repair_costs_chunk, repair_times_chunk = (
        process_buildings_chunk(
            chunk,
            grid_points,
            grid_data,
            sample_size_demand,
            sample_size_damage,
            dl_method,
        )
    )

    # Save each result DataFrame to compressed CSV files
    demand_sample_chunk.to_csv(f'{temp_dir}/demand_part_{i}.csv', compression='zip')
    damage_df_chunk.to_csv(f'{temp_dir}/damage_part_{i}.csv', compression='zip')
    repair_costs_chunk.to_csv(
        f'{temp_dir}/repair_costs_part_{i}.csv', compression='zip'
    )
    repair_times_chunk.to_csv(
        f'{temp_dir}/repair_times_part_{i}.csv', compression='zip'
    )


def regional_sim(config_file: str, num_cores: int | None = None) -> None:
    """
    Perform a regional-scale disaster impact simulation.

    This function orchestrates the complete regional simulation workflow including:
    1. Loading earthquake event data from gridded intensity measure files
    2. Loading building inventory data
    3. Mapping event intensity measures to building locations using nearest neighbor regression
    4. Mapping buildings to damage/loss archetypes using Pelicun auto-population
    5. Calculating damage states for all buildings
    6. Calculating repair costs and times
    7. Aggregating and saving results to CSV files

    The simulation is performed in parallel chunks to handle large building inventories
    efficiently. Results are saved as compressed CSV files for demand samples, damage
    states, repair costs, and repair times.

    Parameters
    ----------
    config_file : str
        Path to JSON configuration file containing simulation parameters,
        file paths, and analysis settings (inputRWHALE.json from SimCenter's R2D Tool)
    num_cores : int, optional
        Number of CPU cores to use for parallel processing. If None,
        uses all available cores minus one

    Notes
    -----
    Output Files:
        - demand_sample.csv: Intensity measure realizations for all buildings
        - damage_sample.csv: Damage state realizations for all building components
        - repair_cost_sample.csv: Repair cost estimates for all buildings
        - repair_time_sample.csv: Repair time estimates for all buildings

    """
    batch_size = 1000

    # Initialize start time for timestamp tracking
    start_time = time.time()

    with Path(config_file).open(encoding='utf-8') as f:
        config = json.load(f)

    sample_size_demand = config['Applications']['RegionalMapping']['Buildings'][
        'ApplicationData'
    ]['samples']
    sample_size_damage = config['Applications']['DL']['Buildings'][
        'ApplicationData'
    ]['Realizations']

    # 1 Earthquake Event
    # Load gridded event IM information from a standard SimCenter EventGrid file and the corresponding site files.

    event_data_folder = config['RegionalEvent']['eventFilePath']
    event_grid_path = f"{event_data_folder}/{config['RegionalEvent']['eventFile']}"

    grid_points = pd.read_csv(event_grid_path)

    grid_point_data_array = []

    for grid_point_file in tqdm(
        grid_points['GP_file'],
        desc=f'[{format_elapsed_time(start_time)}] 1 Earthquake Event - Loading grid point data',
    ):
        grid_point_data = pd.read_csv(
            f'{event_data_folder}/{grid_point_file}', nrows=sample_size_demand
        )

        grid_point_data_array.append(grid_point_data)

    grid_data = pd.concat(grid_point_data_array, axis=1, keys=grid_points.index)

    # 2 Building Inventory
    # Load probabilistic building inventory from a CSV file
    print(f'[{format_elapsed_time(start_time)}] 2 Building Inventory')  # noqa: T201

    bldg_data_folder = config['Applications']['Assets']['Buildings'][
        'ApplicationData'
    ]['pathToSource']
    bldg_data_path = f"{bldg_data_folder}/{config['Applications']['Assets']['Buildings']['ApplicationData']['assetSourceFile']}"

    bldg_df = pd.read_csv(bldg_data_path, index_col=0)
    original_index = bldg_df.index
    bldg_df = bldg_df.sort_values(by='OccupancyClass')

    # bldg_df = bldg_df.iloc[:batch_size*5]

    # Get DL method for processing
    dl_method = config['Applications']['DL']['Buildings']['ApplicationData'][
        'DL_Method'
    ]

    with tempfile.TemporaryDirectory() as temp_dir:
        # Chunk the DataFrame into a list of smaller DataFrames
        bldg_chunks = [
            bldg_df.iloc[i : i + batch_size]
            for i in range(0, len(bldg_df), batch_size)
        ]

        # Determine the number of CPU cores to use
        if num_cores:
            n_jobs = num_cores
        else:
            n_jobs = max(1, os.cpu_count() - 1)

        # Process chunks in parallel with a proper progress bar
        with (
            tqdm(total=len(bldg_chunks), desc='Processing building chunks') as pbar,
            tqdm_joblib(pbar),
        ):
            Parallel(n_jobs=n_jobs)(
                delayed(process_and_save_chunk)(
                    i,
                    chunk,
                    temp_dir,
                    grid_points,
                    grid_data,
                    sample_size_demand,
                    sample_size_damage,
                    dl_method,
                )
                for i, chunk in enumerate(bldg_chunks)
            )

        # Read and combine all temporary files
        demand_list = []
        damage_list = []
        costs_list = []
        times_list = []

        for filename in os.listdir(temp_dir):
            filepath = Path(temp_dir) / filename
            if filename.startswith('demand_part_'):
                demand_list.append(
                    pd.read_csv(filepath, index_col=0, compression='zip')
                )
            elif filename.startswith('damage_part_'):
                damage_list.append(
                    pd.read_csv(filepath, index_col=0, compression='zip')
                )
            elif filename.startswith('repair_costs_part_'):
                costs_list.append(
                    pd.read_csv(filepath, index_col=0, compression='zip')
                )
            elif filename.startswith('repair_times_part_'):
                times_list.append(
                    pd.read_csv(filepath, index_col=0, compression='zip')
                )

        # Concatenate all parts into final DataFrames
        demand_sample = pd.concat(demand_list, axis=1)
        damage_df = pd.concat(damage_list, axis=0)
        repair_costs = pd.concat(costs_list, axis=0)
        repair_times = pd.concat(times_list, axis=0)

        # Restore original building order for all result files
        demand_sample = demand_sample.reindex(
            columns=[f'PGA-{loc}-1' for loc in original_index]
        )
        damage_df = damage_df.reindex(original_index)
        repair_costs = repair_costs.reindex(original_index)
        repair_times = repair_times.reindex(original_index)

        # 7 Save results
        print(f'[{format_elapsed_time(start_time)}] 7 Save results')  # noqa: T201

        demand_sample.to_csv('demand.csv')
        damage_df.to_csv('damage.csv')
        repair_costs.to_csv('repair_costs.csv')
        repair_times.to_csv('repair_times.csv')



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis

"""These are unit and integration tests on the asset model of pelicun."""

from __future__ import annotations

import tempfile
from copy import deepcopy
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pytest

from pelicun import assessment
from pelicun.base import ensure_value
from pelicun.tests.basic.test_pelicun_model import TestPelicunModel

if TYPE_CHECKING:
    from pelicun.model.asset_model import AssetModel


class TestAssetModel(TestPelicunModel):
    @pytest.fixture
    def asset_model(self, assessment_instance: assessment.Assessment) -> AssetModel:
        return deepcopy(assessment_instance.asset)

    def test_init_method(self, asset_model: AssetModel) -> None:
        assert asset_model.log
        assert asset_model.cmp_marginal_params is None
        assert asset_model.cmp_units is None
        assert asset_model._cmp_RVs is None
        assert asset_model.cmp_sample is None

    def test_save_cmp_sample(self, asset_model: AssetModel) -> None:
        asset_model.cmp_sample = pd.DataFrame(
            {
                ('component_a', f'{i}', f'{j}', '0'): 8.0
                for i in range(1, 3)
                for j in range(1, 3)
            },
            index=range(10),
            columns=pd.MultiIndex.from_tuples(
                (
                    ('component_a', f'{i}', f'{j}', '0')
                    for i in range(1, 3)
                    for j in range(1, 3)
                ),
                names=('cmp', 'loc', 'dir', 'uid'),
            ),
        )

        asset_model.cmp_units = pd.Series(
            data=['ea'], index=['component_a'], name='Units'
        )

        res = asset_model.save_cmp_sample()
        assert isinstance(res, pd.DataFrame)

        temp_dir = tempfile.mkdtemp()
        # save the sample there
        asset_model.save_cmp_sample(f'{temp_dir}/temp.csv')

        # load the component sample to a different AssetModel
        asmt = assessment.Assessment()
        asset_model = asmt.asset
        asset_model.load_cmp_sample(f'{temp_dir}/temp.csv')

        # also test loading sample to variables
        # (but we don't inspect them)
        asset_model.save_cmp_sample(save_units=False)
        asset_model.save_cmp_sample(save_units=True)

    def test_load_cmp_model_1(self, asset_model: AssetModel) -> None:
        cmp_marginals = pd.read_csv(
            'pelicun/tests/basic/data/model/test_AssetModel/CMP_marginals.csv',
            index_col=0,
        )
        asset_model.load_cmp_model({'marginals': cmp_marginals})

        expected_cmp_marginal_params = pd.DataFrame(
            {
                'Theta_0': (8.0, 8.0, 8.0, 8.0, 8.0, 8.0),
                'Blocks': (1, 1, 1, 1, 1, 1),
            },
            index=pd.MultiIndex.from_tuples(
                (
                    ('component_a', '0', '1', '0'),
                    ('component_a', '0', '2', '0'),
                    ('component_a', '1', '1', '0'),
                    ('component_a', '1', '2', '0'),
                    ('component_a', '2', '1', '0'),
                    ('component_a', '2', '2', '0'),
                ),
                names=('cmp', 'loc', 'dir', 'uid'),
            ),
        ).astype({'Theta_0': 'float64', 'Blocks': 'int64'})

        pd.testing.assert_frame_equal(
            expected_cmp_marginal_params,
            ensure_value(asset_model.cmp_marginal_params),
            check_index_type=False,
            check_column_type=False,
            check_dtype=False,
        )

        expected_cmp_units = pd.Series(
            data=['ea'], index=['component_a'], name='Units'
        )

        pd.testing.assert_series_equal(
            expected_cmp_units,
            ensure_value(asset_model.cmp_units),
            check_index_type=False,
        )

    def test_load_cmp_model_2(self, asset_model: AssetModel) -> None:
        # component marginals utilizing the keywords '--', 'all', 'top', 'roof'
        cmp_marginals = pd.read_csv(
            'pelicun/tests/basic/data/model/test_AssetModel/CMP_marginals_2.csv',
            index_col=0,
        )
        asset_model._asmnt.stories = 4
        asset_model.load_cmp_model({'marginals': cmp_marginals})

        assert ensure_value(asset_model.cmp_marginal_params).to_dict() == {
            'Theta_0': {
                ('component_a', '0', '1', '0'): 1.0,
                ('component_a', '0', '2', '0'): 1.0,
                ('component_a', '1', '1', '0'): 1.0,
                ('component_a', '1', '2', '0'): 1.0,
                ('component_a', '2', '1', '0'): 1.0,
                ('component_a', '2', '2', '0'): 1.0,
                ('component_a', '3', '1', '0'): 1.0,
                ('component_a', '3', '2', '0'): 1.0,
                ('component_b', '1', '1', '0'): 1.0,
                ('component_b', '2', '1', '0'): 1.0,
                ('component_b', '3', '1', '0'): 1.0,
                ('component_b', '4', '1', '0'): 1.0,
                ('component_c', '0', '1', '0'): 1.0,
                ('component_c', '1', '1', '0'): 1.0,
                ('component_c', '2', '1', '0'): 1.0,
                ('component_d', '4', '1', '0'): 1.0,
                ('component_e', '5', '1', '0'): 1.0,
            },
            'Blocks': {
                ('component_a', '0', '1', '0'): 1,
                ('component_a', '0', '2', '0'): 1,
                ('component_a', '1', '1', '0'): 1,
                ('component_a', '1', '2', '0'): 1,
                ('component_a', '2', '1', '0'): 1,
                ('component_a', '2', '2', '0'): 1,
                ('component_a', '3', '1', '0'): 1,
                ('component_a', '3', '2', '0'): 1,
                ('component_b', '1', '1', '0'): 1,
                ('component_b', '2', '1', '0'): 1,
                ('component_b', '3', '1', '0'): 1,
                ('component_b', '4', '1', '0'): 1,
                ('component_c', '0', '1', '0'): 1,
                ('component_c', '1', '1', '0'): 1,
                ('component_c', '2', '1', '0'): 1,
                ('component_d', '4', '1', '0'): 1,
                ('component_e', '5', '1', '0'): 1,
            },
        }

        expected_cmp_units = pd.Series(
            data=['ea'] * 5,
            index=[f'component_{x}' for x in ('a', 'b', 'c', 'd', 'e')],
            name='Units',
        )

        pd.testing.assert_series_equal(
            expected_cmp_units,
            ensure_value(asset_model.cmp_units),
            check_index_type=False,
        )

    def test_load_cmp_model_csv(self, asset_model: AssetModel) -> None:
        # load by directly specifying the csv file
        cmp_marginals = 'pelicun/tests/basic/data/model/test_AssetModel/CMP'
        asset_model.load_cmp_model(cmp_marginals)

    def test_load_cmp_model_exceptions(self, asset_model: AssetModel) -> None:
        cmp_marginals = pd.read_csv(
            'pelicun/tests/basic/data/model/test_AssetModel/'
            'CMP_marginals_invalid_loc.csv',
            index_col=0,
        )
        asset_model._asmnt.stories = 4
        with pytest.raises(
            ValueError, match='Cannot parse location string: basement'
        ):
            asset_model.load_cmp_model({'marginals': cmp_marginals})

        cmp_marginals = pd.read_csv(
            'pelicun/tests/basic/data/model/test_AssetModel/'
            'CMP_marginals_invalid_dir.csv',
            index_col=0,
        )
        asset_model._asmnt.stories = 4
        with pytest.raises(
            ValueError, match='Cannot parse direction string: non-directional'
        ):
            asset_model.load_cmp_model({'marginals': cmp_marginals})

    def test_generate_cmp_sample(self, asset_model: AssetModel) -> None:
        asset_model.cmp_marginal_params = pd.DataFrame(
            {'Theta_0': (8.0, 8.0, 8.0, 8.0), 'Blocks': (1.0, 1.0, 1.0, 1.0)},
            index=pd.MultiIndex.from_tuples(
                (
                    ('component_a', '1', '1', '0'),
                    ('component_a', '1', '2', '0'),
                    ('component_a', '2', '1', '0'),
                    ('component_a', '2', '2', '0'),
                ),
                names=('cmp', 'loc', 'dir', 'uid'),
            ),
        )

        asset_model.cmp_units = pd.Series(
            data=['ea'], index=['component_a'], name='Units'
        )

        asset_model.generate_cmp_sample(sample_size=10)

        assert asset_model._cmp_RVs is not None

        expected_cmp_sample = pd.DataFrame(
            {
                ('component_a', f'{i}', f'{j}'): 8.0
                for i in range(1, 3)
                for j in range(1, 3)
            },
            index=range(10),
            columns=pd.MultiIndex.from_tuples(
                (
                    ('component_a', f'{i}', f'{j}', '0')
                    for i in range(1, 3)
                    for j in range(1, 3)
                ),
                names=('cmp', 'loc', 'dir', 'uid'),
            ),
        )

        pd.testing.assert_frame_equal(
            expected_cmp_sample,
            ensure_value(asset_model.cmp_sample),
            check_index_type=False,
            check_column_type=False,
        )

    def test_generate_cmp_sample_exceptions_1(self, asset_model: AssetModel) -> None:
        # without marginal parameters
        with pytest.raises(
            ValueError, match='Model parameters have not been specified'
        ):
            asset_model.generate_cmp_sample(sample_size=10)

    def test_generate_cmp_sample_exceptions_2(self, asset_model: AssetModel) -> None:
        # without specifying sample size
        cmp_marginals = pd.read_csv(
            'pelicun/tests/basic/data/model/test_AssetModel/CMP_marginals.csv',
            index_col=0,
        )
        asset_model.load_cmp_model({'marginals': cmp_marginals})
        with pytest.raises(ValueError, match='Sample size was not specified'):
            asset_model.generate_cmp_sample()
        # but it should work if a demand sample is available
        asset_model._asmnt.demand.sample = pd.DataFrame(np.empty(shape=(10, 2)))
        asset_model.generate_cmp_sample()



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis

"""These are unit and integration tests on the assessment module of pelicun."""

from __future__ import annotations

import pytest

from pelicun import assessment


def create_assessment_obj(config: dict | None = None) -> assessment.Assessment:
    return assessment.Assessment(config) if config else assessment.Assessment({})


def test_Assessment_init() -> None:
    asmt = create_assessment_obj()
    # confirm attributes
    for attribute in (
        'asset',
        'calc_unit_scale_factor',
        'damage',
        'demand',
        'get_default_data',
        'get_default_metadata',
        'log',
        'loss',
        'options',
        'scale_factor',
        'stories',
        'unit_conversion_factors',
    ):
        assert hasattr(asmt, attribute)
    # confirm that creating an attribute on the fly is not allowed
    with pytest.raises(AttributeError):
        asmt.my_attribute = 2  # type: ignore


def test_assessment_get_default_metadata() -> None:
    asmt = create_assessment_obj()

    method_names = (
        # test for backwards compatibility
        'damage_DB_FEMA_P58_2nd',
        'damage_DB_Hazus_EQ_bldg',
        'damage_DB_Hazus_EQ_trnsp',
        'loss_repair_DB_FEMA_P58_2nd',
        'loss_repair_DB_Hazus_EQ_bldg',
        'loss_repair_DB_Hazus_EQ_trnsp',
        # current valid values
        'Hazus Earthquake - Buildings',
        'Hazus Earthquake - Stories',
        'Hazus Earthquake - Transportation',
        'Hazus Hurricane Wind - Buildings',
    )

    for method_name in method_names:
        for model_type in ['fragility', 'consequence_repair']:
            if method_name.startswith(('damage', 'loss')):
                model_type = None  # noqa: PLW2901

            # here we just test that we can load the data file, without
            # checking the contents.
            asmt.get_default_data(method_name, model_type)
            asmt.get_default_metadata(method_name, model_type)


def test_assessment_calc_unit_scale_factor() -> None:
    # default unit file
    asmt = create_assessment_obj()

    # without specifying a quantity
    assert asmt.calc_unit_scale_factor('m') == 1.00
    assert asmt.calc_unit_scale_factor('in') == 0.0254

    # with quantity
    assert asmt.calc_unit_scale_factor('2.00 m') == 2.00
    assert asmt.calc_unit_scale_factor('2 in') == 2.00 * 0.0254

    # when a custom unit file is specified, changing the base units
    asmt = create_assessment_obj(
        {
            'UnitsFile': (
                'pelicun/tests/basic/data/assessment/'
                'test_assessment_calc_unit_scale_factor/'
                'custom_units.json'
            )
        }
    )

    assert asmt.calc_unit_scale_factor('in') == 1.00
    assert asmt.calc_unit_scale_factor('m') == 39.3701

    # exceptions

    # unrecognized unit
    with pytest.raises(KeyError):
        asmt.calc_unit_scale_factor('smoot')
        # 1 smoot was 67 inches in 1958.


def test_assessment_scale_factor() -> None:
    # default unit file
    asmt = create_assessment_obj()
    assert asmt.scale_factor('m') == 1.00
    assert asmt.scale_factor('in') == 0.0254

    # when a custom unit file is specified, changing the base units
    asmt = create_assessment_obj(
        {
            'UnitsFile': (
                'pelicun/tests/basic/data/assessment/'
                'test_assessment_calc_unit_scale_factor/'
                'custom_units.json'
            )
        }
    )

    assert asmt.scale_factor('in') == 1.00
    assert asmt.scale_factor('m') == 39.3701

    # exceptions
    with pytest.raises(ValueError, match='Unknown unit: helen'):
        asmt.scale_factor('helen')



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.

"""
Validation test for the probability of each damage state of a
component.

"""

from __future__ import annotations

import tempfile

import numpy as np
import pandas as pd
from scipy.stats import norm  # type: ignore

from pelicun import assessment, file_io


def test_validation_ds_probabilities() -> None:
    sample_size = 1000000

    asmnt = assessment.Assessment({'PrintLog': False, 'Seed': 42})

    #
    # Demands
    #

    demands = pd.DataFrame(
        {
            'Theta_0': [0.015],
            'Theta_1': [0.60],
            'Family': ['lognormal'],
            'Units': ['rad'],
        },
        index=pd.MultiIndex.from_tuples(
            [
                ('PID', '1', '1'),
            ],
        ),
    )

    # load the demand model
    asmnt.demand.load_model({'marginals': demands})

    # generate samples
    asmnt.demand.generate_sample({'SampleSize': sample_size})

    #
    # Asset
    #

    # specify number of stories
    asmnt.stories = 1

    # load component definitions
    cmp_marginals = pd.read_csv(
        'pelicun/tests/validation/v1/data/CMP_marginals.csv', index_col=0
    )
    cmp_marginals['Blocks'] = cmp_marginals['Blocks']
    asmnt.asset.load_cmp_model({'marginals': cmp_marginals})

    # generate sample
    asmnt.asset.generate_cmp_sample(sample_size)

    #
    # Damage
    #

    damage_db = file_io.load_data(
        'pelicun/tests/validation/v1/data/damage_db.csv',
        reindex=False,
        unit_conversion_factors=asmnt.unit_conversion_factors,
    )
    assert isinstance(damage_db, pd.DataFrame)

    cmp_set = set(asmnt.asset.list_unique_component_ids())

    # load the models into pelicun
    asmnt.damage.load_model_parameters([damage_db], cmp_set)

    # calculate damages
    asmnt.damage.calculate()

    probs = asmnt.damage.ds_model.probabilities()

    #
    # Analytical calculation of the probability of each damage state
    #

    demand_median = 0.015
    demand_beta = 0.60
    capacity_1_median = 0.015
    capacity_2_median = 0.02
    capacity_beta = 0.50

    # If Y is LogNormal(delta, beta), then X = Log(Y) is Normal(mu, sigma)
    # with mu = log(delta) and sigma = beta
    demand_mean = np.log(demand_median)
    capacity_1_mean = np.log(capacity_1_median)
    capacity_2_mean = np.log(capacity_2_median)
    demand_std = demand_beta
    capacity_std = capacity_beta

    p0 = 1.00 - norm.cdf(
        (demand_mean - capacity_1_mean) / np.sqrt(demand_std**2 + capacity_std**2)
    )
    p1 = norm.cdf(
        (demand_mean - capacity_1_mean) / np.sqrt(demand_std**2 + capacity_std**2)
    ) - norm.cdf(
        (demand_mean - capacity_2_mean) / np.sqrt(demand_std**2 + capacity_std**2)
    )
    p2 = norm.cdf(
        (demand_mean - capacity_2_mean) / np.sqrt(demand_std**2 + capacity_std**2)
    )

    assert np.allclose(probs.iloc[0, 0], p0, atol=1e-2)  # type: ignore
    assert np.allclose(probs.iloc[0, 1], p1, atol=1e-2)  # type: ignore
    assert np.allclose(probs.iloc[0, 2], p2, atol=1e-2)  # type: ignore

    #
    # Also test load/save sample
    #

    assert asmnt.damage.ds_model.sample is not None
    asmnt.damage.ds_model.sample = asmnt.damage.ds_model.sample.iloc[0:100, :]
    # (we reduce the number of realizations to conserve resources)
    before = asmnt.damage.ds_model.sample.copy()
    temp_dir = tempfile.mkdtemp()
    asmnt.damage.save_sample(f'{temp_dir}/mdl.csv')
    asmnt.damage.load_sample(f'{temp_dir}/mdl.csv')
    pd.testing.assert_frame_equal(before, asmnt.damage.ds_model.sample)



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis

"""This file defines a class used by the model unit tests."""

from __future__ import annotations

from copy import deepcopy
from typing import Callable

import pytest

from pelicun import assessment


class TestModelModule:
    @pytest.fixture
    def assessment_factory(self) -> Callable:
        def create_instance(*, verbose: bool) -> assessment.Assessment:
            x = assessment.Assessment()
            x.log.verbose = verbose
            return x

        return create_instance

    @pytest.fixture(params=[True, False])
    def assessment_instance(self, request, assessment_factory) -> None:  # noqa: ANN001
        return deepcopy(assessment_factory(verbose=request.param))



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.

from __future__ import annotations

from typing import Generator
from unittest.mock import patch

import pytest

# This is an explicit list of tests that handle their own network mocking
# and should be excluded from the global fixture.
EXCLUDED_TESTS = {
    'test_check_dlml_data_with_missing_data',
    'test_check_dlml_data_with_existing_data_update_available',
    'test_check_dlml_data_download_failure',
    'test_check_dlml_data_permission_error',
    'test_check_dlml_data_version_check_failure',
    'test_logging_configuration',
    'test_warning_system_integration',
}


@pytest.fixture(autouse=True)
def mock_dlml_data_check(
    request: pytest.FixtureRequest,
) -> Generator[None, None, None]:
    """
    Mocks the DLML data check for the entire test session, UNLESS
    the test is in the specific exclusion list.

    This provides a high-performance, fail-safe "no network" policy
    while allowing specific integration tests to run their own logic.
    """
    # If the current test's name is in our exclusion list, do not
    # apply the global mock.
    if request.node.name in EXCLUDED_TESTS:
        yield
    else:
        # For all other tests, apply the global mock.
        target_function_path = 'pelicun.tools.dlml.check_dlml_data'
        with patch(target_function_path) as mocked_check_func:
            mocked_check_func.return_value = None
            yield
